import sempy.fabric as fabric
import pandas as pd
import warnings
import datetime
from IPython.display import display, HTML
from sempy_labs._model_dependencies import get_model_calc_dependencies
from sempy_labs._helper_functions import (
    format_dax_object_name,
    create_relationship_name,
    save_as_delta_table,
    resolve_workspace_capacity,
    resolve_dataset_name_and_id,
    get_language_codes,
    _get_column_aggregate,
    resolve_workspace_name_and_id,
    _create_spark_session,
)
from sempy_labs.lakehouse import get_lakehouse_tables, lakehouse_attached
from sempy_labs.tom import connect_semantic_model
from sempy_labs._model_bpa_rules import model_bpa_rules
from typing import Optional
from sempy._utils._log import log
import sempy_labs._icons as icons
from pyspark.sql.functions import col, flatten
from pyspark.sql.types import StructType, StructField, StringType
import os
from uuid import UUID


@log
def run_model_bpa(
    dataset: str | UUID,
    rules: Optional[pd.DataFrame] = None,
    workspace: Optional[str] = None,
    export: bool = False,
    return_dataframe: bool = False,
    extended: bool = False,
    language: Optional[str] = None,
    check_dependencies: bool = True,
    **kwargs,
):
    """
    Displays an HTML visualization of the results of the Best Practice Analyzer scan for a semantic model.

    The Best Practice Analyzer rules are based on the rules defined `here <https://github.com/microsoft/Analysis-Services/tree/master/BestPracticeRules>`_. The framework for the Best Practice Analyzer and rules are based on the foundation set by `Tabular Editor <https://github.com/TabularEditor/TabularEditor>`_.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    rules : pandas.DataFrame, default=None
        A pandas dataframe containing rules to be evaluated.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    export : bool, default=False
        If True, exports the resulting dataframe to a delta table in the lakehouse attached to the notebook.
    return_dataframe : bool, default=False
        If True, returns a pandas dataframe instead of the visualization.
    extended : bool, default=False
        If True, runs the set_vertipaq_annotations function to collect Vertipaq Analyzer statistics to be used in the analysis of the semantic model.
    language : str, default=None
        Specifying a language name or code (i.e. 'it-IT' for Italian) will auto-translate the Category, Rule Name and Description into the specified language.
        Defaults to None which resolves to English.
    check_dependencies : bool, default=True
        If True, leverages the model dependencies from get_model_calc_dependencies to evaluate the rules. Set this parameter to False if running the rules against a semantic model in a shared capacity.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe in HTML format showing semantic model objects which violated the best practice analyzer rules.
    """

    import polib

    if "extend" in kwargs:
        print(
            "The 'extend' parameter has been deprecated. Please remove this parameter from the function going forward."
        )
        del kwargs["extend"]

    warnings.filterwarnings(
        "ignore",
        message="This pattern is interpreted as a regular expression, and has match groups.",
    )
    warnings.filterwarnings(
        "ignore", category=UserWarning, message=".*Arrow optimization.*"
    )

    language_list = list(icons.language_map.keys())
    if language is not None:
        language = get_language_codes(languages=language)[0]

    # Map languages to the closest language (first 2 letters matching)
    def map_language(language, language_list):

        mapped = False

        if language in language_list:
            mapped is True
            return language

        language_prefix = language[:2]
        for lang_code in language_list:
            if lang_code.startswith(language_prefix):
                mapped is True
                return lang_code
        if not mapped:
            return language

        if language is not None:
            language = map_language(language, language_list)

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(
        dataset, workspace=workspace_id
    )

    if language is not None and language not in language_list:
        print(
            f"{icons.yellow_dot} The '{language}' language code is not in our predefined language list. Please file an issue and let us know which language code you are using: https://github.com/microsoft/semantic-link-labs/issues/new?assignees=&labels=&projects=&template=bug_report.md&title=."
        )

    with connect_semantic_model(
        dataset=dataset_id, workspace=workspace_id, readonly=True
    ) as tom:

        if extended:
            tom.set_vertipaq_annotations()

        # Do not run BPA for models with no tables
        if tom.model.Tables.Count == 0:
            print(
                f"{icons.warning} The '{dataset_name}' semantic model within the '{workspace_name}' workspace has no tables and therefore there are no valid BPA results."
            )
            return

        if check_dependencies:
            dep = get_model_calc_dependencies(
                dataset=dataset_id, workspace=workspace_id
            )
        else:
            dep = pd.DataFrame(
                columns=[
                    "Table Name",
                    "Object Name",
                    "Object Type",
                    "Expression",
                    "Referenced Table",
                    "Referenced Object",
                    "Referenced Object Type",
                    "Full Object Name",
                    "Referenced Full Object Name",
                    "Parent Node",
                ]
            )

        def translate_using_po(rule_file):
            current_dir = os.path.dirname(os.path.abspath(__file__))
            translation_file = (
                f"{current_dir}/_bpa_translation/_model/_translations_{language}.po"
            )
            for c in ["Category", "Description", "Rule Name"]:
                po = polib.pofile(translation_file)
                for entry in po:
                    if entry.tcomment == c.lower().replace(" ", "_"):
                        rule_file.loc[rule_file["Rule Name"] == entry.msgid, c] = (
                            entry.msgstr
                        )

        translated = False

        # Translations
        if language is not None and rules is None and language in language_list:
            rules = model_bpa_rules(dependencies=dep)
            translate_using_po(rules)
            translated = True
        if rules is None:
            rules = model_bpa_rules(dependencies=dep)
        if language is not None and not translated:

            def translate_using_spark(rule_file):

                from synapse.ml.services import Translate

                rules_temp = rule_file.copy()
                rules_temp = rules_temp.drop(["Expression", "URL", "Severity"], axis=1)

                schema = StructType(
                    [
                        StructField("Category", StringType(), True),
                        StructField("Scope", StringType(), True),
                        StructField("Rule Name", StringType(), True),
                        StructField("Description", StringType(), True),
                    ]
                )

                spark = _create_spark_session()
                dfRules = spark.createDataFrame(rules_temp, schema)

                columns = ["Category", "Rule Name", "Description"]
                for clm in columns:
                    translate = (
                        Translate()
                        .setTextCol(clm)
                        .setToLanguage(language)
                        .setOutputCol("translation")
                        .setConcurrency(5)
                    )

                    if clm == "Rule Name":
                        transDF = (
                            translate.transform(dfRules)
                            .withColumn(
                                "translation",
                                flatten(col("translation.translations")),
                            )
                            .withColumn("translation", col("translation.text"))
                            .select(clm, "translation")
                        )
                    else:
                        transDF = (
                            translate.transform(dfRules)
                            .withColumn(
                                "translation",
                                flatten(col("translation.translations")),
                            )
                            .withColumn("translation", col("translation.text"))
                            .select("Rule Name", clm, "translation")
                        )

                    df_panda = transDF.toPandas()
                    rule_file = pd.merge(
                        rule_file,
                        df_panda[["Rule Name", "translation"]],
                        on="Rule Name",
                        how="left",
                    )

                    rule_file = rule_file.rename(
                        columns={"translation": f"{clm}Translated"}
                    )
                    rule_file[f"{clm}Translated"] = rule_file[f"{clm}Translated"].apply(
                        lambda x: x[0] if x is not None else None
                    )

                for clm in columns:
                    rule_file = rule_file.drop([clm], axis=1)
                    rule_file = rule_file.rename(columns={f"{clm}Translated": clm})

                return rule_file

            rules = translate_using_spark(rules)

        rules.loc[rules["Severity"] == "Warning", "Severity"] = icons.warning
        rules.loc[rules["Severity"] == "Error", "Severity"] = icons.error
        rules.loc[rules["Severity"] == "Info", "Severity"] = icons.info

        pd.set_option("display.max_colwidth", 1000)

        violations = pd.DataFrame(columns=["Object Name", "Scope", "Rule Name"])

        scope_to_dataframe = {
            "Relationship": (
                tom.model.Relationships,
                lambda obj: create_relationship_name(
                    obj.FromTable.Name,
                    obj.FromColumn.Name,
                    obj.ToTable.Name,
                    obj.ToColumn.Name,
                ),
            ),
            "Column": (
                tom.all_columns(),
                lambda obj: format_dax_object_name(obj.Parent.Name, obj.Name),
            ),
            "Calculated Column": (
                tom.all_calculated_columns(),
                lambda obj: format_dax_object_name(obj.Parent.Name, obj.Name),
            ),
            "Measure": (tom.all_measures(), lambda obj: obj.Name),
            "Hierarchy": (
                tom.all_hierarchies(),
                lambda obj: format_dax_object_name(obj.Parent.Name, obj.Name),
            ),
            "Table": (tom.model.Tables, lambda obj: obj.Name),
            "Calculated Table": (tom.all_calculated_tables(), lambda obj: obj.Name),
            "Role": (tom.model.Roles, lambda obj: obj.Name),
            "Model": (tom.model, lambda obj: obj.Model.Name),
            "Calculation Item": (
                tom.all_calculation_items(),
                lambda obj: format_dax_object_name(obj.Parent.Table.Name, obj.Name),
            ),
            "Row Level Security": (
                tom.all_rls(),
                lambda obj: format_dax_object_name(obj.Parent.Name, obj.Name),
            ),
            "Partition": (
                tom.all_partitions(),
                lambda obj: format_dax_object_name(obj.Parent.Name, obj.Name),
            ),
            "Function": (
                tom.all_functions(),
                lambda obj: obj.Name,
            ),
        }

        for i, r in rules.iterrows():
            ruleName = r["Rule Name"]
            expr = r["Expression"]
            scopes = r["Scope"]

            if isinstance(scopes, str):
                scopes = [scopes]

            for scope in scopes:
                func = scope_to_dataframe[scope][0]
                nm = scope_to_dataframe[scope][1]

                if scope == "Model":
                    x = []
                    if expr(func, tom):
                        x = ["Model"]
                elif scope == "Measure":
                    x = [nm(obj) for obj in tom.all_measures() if expr(obj, tom)]
                elif scope == "Function":
                    x = [nm(obj) for obj in tom.all_functions() if expr(obj, tom)]
                elif scope == "Column":
                    x = [nm(obj) for obj in tom.all_columns() if expr(obj, tom)]
                elif scope == "Partition":
                    x = [nm(obj) for obj in tom.all_partitions() if expr(obj, tom)]
                elif scope == "Hierarchy":
                    x = [nm(obj) for obj in tom.all_hierarchies() if expr(obj, tom)]
                elif scope == "Table":
                    x = [nm(obj) for obj in tom.model.Tables if expr(obj, tom)]
                elif scope == "Calculated Table":
                    x = [
                        nm(obj) for obj in tom.all_calculated_tables() if expr(obj, tom)
                    ]
                elif scope == "Relationship":
                    x = [nm(obj) for obj in tom.model.Relationships if expr(obj, tom)]
                elif scope == "Role":
                    x = [nm(obj) for obj in tom.model.Roles if expr(obj, tom)]
                elif scope == "Row Level Security":
                    x = [nm(obj) for obj in tom.all_rls() if expr(obj, tom)]
                elif scope == "Calculation Item":
                    x = [
                        nm(obj) for obj in tom.all_calculation_items() if expr(obj, tom)
                    ]
                elif scope == "Calculated Column":
                    x = [
                        nm(obj)
                        for obj in tom.all_calculated_columns()
                        if expr(obj, tom)
                    ]

                if len(x) > 0:
                    new_data = {
                        "Object Name": x,
                        "Scope": scope,
                        "Rule Name": ruleName,
                    }
                    violations = pd.concat(
                        [violations, pd.DataFrame(new_data)], ignore_index=True
                    )

        prepDF = pd.merge(
            violations,
            rules[["Rule Name", "Category", "Severity", "Description", "URL"]],
            left_on="Rule Name",
            right_on="Rule Name",
            how="left",
        )
        prepDF.rename(columns={"Scope": "Object Type"}, inplace=True)
        finalDF = prepDF[
            [
                "Category",
                "Rule Name",
                "Severity",
                "Object Type",
                "Object Name",
                "Description",
                "URL",
            ]
        ]

    if export:
        if not lakehouse_attached():
            raise ValueError(
                f"{icons.red_dot} In order to save the Best Practice Analyzer results, a lakehouse must be attached to the notebook. Please attach a lakehouse to this notebook."
            )

        dfExport = finalDF.copy()
        delta_table_name = "modelbparesults"

        lakeT = get_lakehouse_tables()
        lakeT_filt = lakeT[lakeT["Table Name"] == delta_table_name]

        dfExport["Severity"] = dfExport["Severity"].replace(icons.severity_mapping)

        if len(lakeT_filt) == 0:
            runId = 1
        else:
            max_run_id = _get_column_aggregate(table_name=delta_table_name)
            runId = max_run_id + 1

        now = datetime.datetime.now()
        dfD = fabric.list_datasets(workspace=workspace_id, mode="rest")
        dfD_filt = dfD[dfD["Dataset Id"] == dataset_id]
        configured_by = dfD_filt["Configured By"].iloc[0]
        capacity_id, capacity_name = resolve_workspace_capacity(workspace=workspace_id)
        dfExport["Capacity Name"] = capacity_name
        dfExport["Capacity Id"] = capacity_id
        dfExport["Workspace Name"] = workspace_name
        dfExport["Workspace Id"] = workspace_id
        dfExport["Dataset Name"] = dataset_name
        dfExport["Dataset Id"] = dataset_id
        dfExport["Configured By"] = configured_by
        dfExport["Timestamp"] = now
        dfExport["RunId"] = runId
        dfExport["RunId"] = dfExport["RunId"].astype("int")

        dfExport = dfExport[list(icons.bpa_schema.keys())]
        dfExport["RunId"] = dfExport["RunId"].astype("int")
        schema = {
            key.replace(" ", "_"): value for key, value in icons.bpa_schema.items()
        }
        save_as_delta_table(
            dataframe=dfExport,
            delta_table_name=delta_table_name,
            write_mode="append",
            schema=schema,
            merge_schema=True,
        )

    if return_dataframe:
        return finalDF

    pd.set_option("display.max_colwidth", 100)

    finalDF = (
        finalDF[
            [
                "Category",
                "Rule Name",
                "Object Type",
                "Object Name",
                "Severity",
                "Description",
                "URL",
            ]
        ]
        .sort_values(["Category", "Rule Name", "Object Type", "Object Name"])
        .set_index(["Category", "Rule Name"])
    )

    bpa2 = finalDF.reset_index()
    bpa_dict = {
        cat: bpa2[bpa2["Category"] == cat].drop("Category", axis=1)
        for cat in bpa2["Category"].drop_duplicates().values
    }

    styles = """
    <style>
        .tab { overflow: hidden; border: 1px solid #ccc; background-color: #f1f1f1; }
        .tab button { background-color: inherit; float: left; border: none; outline: none; cursor: pointer; padding: 14px 16px; transition: 0.3s; }
        .tab button:hover { background-color: #ddd; }
        .tab button.active { background-color: #ccc; }
        .tabcontent { display: none; padding: 6px 12px; border: 1px solid #ccc; border-top: none; }
        .tabcontent.active { display: block; }
        .tooltip { position: relative; display: inline-block; }
        .tooltip .tooltiptext { visibility: hidden; width: 300px; background-color: #555; color: #fff; text-align: center; border-radius: 6px; padding: 5px; position: absolute; z-index: 1; bottom: 125%; left: 50%; margin-left: -110px; opacity: 0; transition: opacity 0.3s; }
        .tooltip:hover .tooltiptext { visibility: visible; opacity: 1; }
    </style>
    """

    # JavaScript for tab functionality
    script = """
    <script>
    function openTab(evt, tabName) {
        var i, tabcontent, tablinks;
        tabcontent = document.getElementsByClassName("tabcontent");
        for (i = 0; i < tabcontent.length; i++) {
            tabcontent[i].style.display = "none";
        }
        tablinks = document.getElementsByClassName("tablinks");
        for (i = 0; i < tablinks.length; i++) {
            tablinks[i].className = tablinks[i].className.replace(" active", "");
        }
        document.getElementById(tabName).style.display = "block";
        evt.currentTarget.className += " active";
    }
    </script>
    """

    # JavaScript for dynamic tooltip positioning
    dynamic_script = """
    <script>
    function adjustTooltipPosition(event) {
        var tooltip = event.target.querySelector('.tooltiptext');
        var rect = tooltip.getBoundingClientRect();
        var topSpace = rect.top;
        var bottomSpace = window.innerHeight - rect.bottom;

        if (topSpace < bottomSpace) {
            tooltip.style.bottom = '125%';
        } else {
            tooltip.style.bottom = 'auto';
            tooltip.style.top = '125%';
        }
    }
    </script>
    """

    # HTML for tabs
    tab_html = '<div class="tab">'
    content_html = ""
    for i, (title, df) in enumerate(bpa_dict.items()):
        if df.shape[0] == 0:
            continue

        tab_id = f"tab{i}"
        active_class = ""
        if i == 0:
            active_class = "active"

        summary = " + ".join(
            [f"{idx} ({v})" for idx, v in df["Severity"].value_counts().items()]
        )
        tab_html += f'<button class="tablinks {active_class}" onclick="openTab(event, \'{tab_id}\')"><b>{title}</b><br/>{summary}</button>'
        content_html += f'<div id="{tab_id}" class="tabcontent {active_class}">'

        # Adding tooltip for Rule Name using Description column
        content_html += '<table border="1">'
        content_html += "<tr><th>Rule Name</th><th>Object Type</th><th>Object Name</th><th>Severity</th></tr>"
        for _, row in df.iterrows():
            content_html += "<tr>"
            if pd.notnull(row["URL"]):
                content_html += f'<td class="tooltip" onmouseover="adjustTooltipPosition(event)"><a href="{row["URL"]}">{row["Rule Name"]}</a><span class="tooltiptext">{row["Description"]}</span></td>'
            elif pd.notnull(row["Description"]):
                content_html += f'<td class="tooltip" onmouseover="adjustTooltipPosition(event)">{row["Rule Name"]}<span class="tooltiptext">{row["Description"]}</span></td>'
            else:
                content_html += f'<td>{row["Rule Name"]}</td>'
            content_html += f'<td>{row["Object Type"]}</td>'
            content_html += f'<td>{row["Object Name"]}</td>'
            content_html += f'<td style="text-align: center;">{row["Severity"]}</td>'
            # content_html += f'<td>{row["Severity"]}</td>'
            content_html += "</tr>"
        content_html += "</table>"

        content_html += "</div>"
    tab_html += "</div>"

    # Display the tabs, tab contents, and run the script
    if not export:
        return display(HTML(styles + tab_html + content_html + script))



from sempy_labs._helper_functions import (
    resolve_item_id,
    _base_api,
    resolve_workspace_name_and_id,
)
from typing import Optional
from uuid import UUID
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def start_mirroring(sql_database: str | UUID, workspace: Optional[str | UUID] = None):
    """
    Starts data mirroring for the specified SQL Database.

    This is a wrapper function for the following API: `Mirroring - Start Mirroring <https://learn.microsoft.com/rest/api/fabric/sqldatabase/mirroring/start-mirroring>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    sql_database : str | uuid.UUID
        Name or ID of the SQL Database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = resolve_item_id(
        item=sql_database, type="SQLDatabase", workspace=workspace_id
    )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/sqlDatabases/{item_id}/startMirroring",
        method="post",
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The SQL Database '{sql_database}' in the '{workspace_name}' workspace is now being mirrored."
    )


@log
def stop_mirroring(sql_database: str | UUID, workspace: Optional[str | UUID] = None):
    """
    Stops data mirroring for the specified SQL Database.

    This is a wrapper function for the following API: `Mirroring - Stop Mirroring <https://learn.microsoft.com/rest/api/fabric/sqldatabase/mirroring/stop-mirroring>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    sql_database : str | uuid.UUID
        Name or ID of the SQL Database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = resolve_item_id(
        item=sql_database, type="SQLDatabase", workspace=workspace_id
    )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/sqlDatabases/{item_id}/stopMirroring",
        method="post",
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The SQL Database '{sql_database}' in the '{workspace_name}' workspace is no longer being mirrored."
    )



from sempy_labs._helper_functions import (
    resolve_item_id,
    resolve_workspace_id,
    resolve_workspace_name_and_id,
    resolve_item_name_and_id,
    _base_api,
)
from typing import Optional, List
from sempy._utils._log import log
import sempy_labs._icons as icons
from uuid import UUID
import sempy.fabric as fabric


@log
def report_rebind(
    report: str | UUID | List[str | UUID],
    dataset: str | UUID,
    report_workspace: Optional[str | UUID] = None,
    dataset_workspace: Optional[str | UUID] = None,
):
    """
    Rebinds a report to a semantic model.

    This is a wrapper function for the following API: `Reports - Rebind Report In Group <https://learn.microsoft.com/rest/api/power-bi/reports/rebind-report-in-group>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    report : str | uuid.UUID | List[str | uuid.UUID]
        Name(s) or ID(s) of the Power BI report(s).
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    report_workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace in which the report resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    dataset_workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace in which the semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (report_workspace_name, report_workspace_id) = resolve_workspace_name_and_id(
        report_workspace
    )

    if dataset_workspace is None:
        dataset_workspace = report_workspace_name

    (dataset_workspace_name, dataset_workspace_id) = resolve_workspace_name_and_id(
        dataset_workspace
    )
    if isinstance(report, str):
        report = [report]

    for rpt in report:
        (report_name, report_id) = resolve_item_name_and_id(
            item=rpt, type="Report", workspace=report_workspace_id
        )
        (dataset_name, dataset_id) = resolve_item_name_and_id(
            item=dataset, type="SemanticModel", workspace=dataset_workspace
        )

        payload = {"datasetId": dataset_id}

        _base_api(
            request=f"v1.0/myorg/groups/{report_workspace_id}/reports/{report_id}/Rebind",
            method="post",
            payload=payload,
            client="fabric_sp",
        )

        print(
            f"{icons.green_dot} The '{report_name}' report within the '{report_workspace_name}' workspace has been successfully rebinded to the '{dataset_name}' semantic model within the '{dataset_workspace_name}' workspace."
        )


@log
def report_rebind_all(
    dataset: str | UUID,
    new_dataset: str | UUID,
    dataset_workspace: Optional[str | UUID] = None,
    new_dataset_workspace: Optional[str | UUID] = None,
    report_workspace: Optional[str | UUID | List[str | UUID]] = None,
):
    """
    Rebinds all reports across the provided report workspaces which are bound to a specific semantic model to a new semantic model.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name of the semantic model currently binded to the reports.
    new_dataset : str | uuid.UUID
        Name of the semantic model to rebind to the reports.
    dataset_workspace : str | uuid.UUID, default=None
        The name of the Fabric workspace in which the original semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    new_dataset_workspace : str | uuid.UUID, default=None
        The name of the Fabric workspace in which the new semantic model resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    report_workspace : str | uuid.UUID | List[str | uuid.UUID], default=None
        The name(s) or IDs of the Fabric workspace(s) in which the report(s) reside(s).
        Defaults to None which finds all reports in all workspaces which use the semantic model and rebinds them to
        the new semantic model.
    """

    (dataset_name, dataset_id) = resolve_item_name_and_id(
        item=dataset, type="SemanticModel", workspace=dataset_workspace
    )
    new_dataset_id = resolve_item_id(
        item=new_dataset, type="SemanticModel", workspace=new_dataset_workspace
    )

    if dataset_id == new_dataset_id:
        raise ValueError(
            f"{icons.red_dot} The 'dataset' and 'new_dataset' parameters are both set to the same semantic model within the same workspace. These parameters must be set to different values."
        )
    dataset_workspace_id = resolve_workspace_id(workspace=dataset_workspace)

    if isinstance(report_workspace, str) or report_workspace is None:
        report_workspace = [report_workspace]

    for w in report_workspace:
        dfR = fabric.list_reports(workspace=w)
        dfR_filt = dfR[
            (dfR["Dataset ID"] == dataset_id)
            & (dfR["Dataset Workspace Id"] == dataset_workspace_id)
        ]
        if dfR_filt.empty:
            (wksp_name, _) = resolve_workspace_name_and_id(workspace=w)
            print(
                f"{icons.info} No reports found for the '{dataset_name}' semantic model within the '{wksp_name}' workspace."
            )
        else:
            # Rebind reports to new dataset
            for _, r in dfR_filt.iterrows():
                rpt_name = r["Name"]
                report_rebind(
                    report=rpt_name,
                    dataset=new_dataset,
                    report_workspace=w,
                    dataset_workspace=new_dataset_workspace,
                )



import pandas as pd
from uuid import UUID
from sempy._utils._log import log
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
    _update_dataframe_datatypes,
)


@log
def list_teams() -> pd.DataFrame:
    """
    Shows a list of teams and their properties.

    This is a wrapper function for the following API: `List teams <https://learn.microsoft.com/graph/api/teams-list>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of teams and their properties.
    """

    result = _base_api(request="teams", client="graph", uses_pagination=True)

    columns = {
        "Team Id": "str",
        "Team Name": "str",
        "Description": "str",
        "Creation Date Time": "datetime",
        "Classification": "str",
        "Specialization": "str",
        "Visibility": "str",
        "Web Url": "str",
        "Archived": "bool",
        "Favorite By Me": "bool",
        "Discoverable By Me": "bool",
        "Member Count": "int_fillna",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    for r in result:
        for v in r.get("value", []):
            rows.append(
                {
                    "Team Id": v.get("id"),
                    "Team Name": v.get("displayName"),
                    "Description": v.get("description"),
                    "Creation Date Time": v.get("createdDateTime"),
                    "Classification": v.get("classification"),
                    "Specialization": v.get("specialization"),
                    "Visibility": v.get("visibility"),
                    "Web Url": v.get("webUrl"),
                    "Archived": v.get("isArchived"),
                    "Favorite By Me": v.get("isFavoriteByMe"),
                    "Discoverable By Me": v.get("isDiscoverableByMe"),
                    "Member Count": v.get("memberCount"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def list_chats(user: str | UUID) -> pd.DataFrame:
    """
    In progress...
    """

    from sempy_labs.graph._users import resolve_user_id

    user_id = resolve_user_id(user=user)
    result = _base_api(request=f"users/{user_id}/chats", client="graph").json()

    columns = {
        "Chat Id": "str",
        "Type": "str",
        "Members": "str",
    }

    df = _create_dataframe(columns=columns)

    rows = []
    for v in result.get("value"):
        rows.append(
            {
                "Chat Id": v.get("id"),
                "Type": v.get("chatType"),
                "Members": v.get("members"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def send_teams_message(chat_id: str, message: str):
    """
    In progress...
    """

    payload = {
        "body": {
            "content": message,
        }
    }

    _base_api(
        request=f"chats/{chat_id}/messages",
        client="graph",
        method="post",
        payload=payload,
        status_codes=201,
    )



import pandas as pd
import sempy_labs._icons as icons
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_workspace_id,
    _base_api,
    _create_dataframe,
    resolve_item_id,
    delete_item,
    create_item,
)
from uuid import UUID
from sempy._utils._log import log


@log
def create_environment(
    environment: str,
    description: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Creates a Fabric environment.

    This is a wrapper function for the following API: `Items - Create Environment <https://learn.microsoft.com/rest/api/fabric/environment/items/create-environment>`_.

    Parameters
    ----------
    environment: str
        Name of the environment.
    description : str, default=None
        A description of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    create_item(
        name=environment,
        description=description,
        type="Environment",
        workspace=workspace,
    )


@log
def list_environments(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the environments within a workspace.

    This is a wrapper function for the following API: `Items - List Environments <https://learn.microsoft.com/rest/api/fabric/environment/items/list-environments>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the environments within a workspace.
    """

    columns = {
        "Environment Name": "string",
        "Environment Id": "string",
        "Description": "string",
        "Publish State": "string",
        "Publish Target Version": "string",
        "Publish Start Time": "string",
        "Publish End Time": "string",
        "Spark Libraries State": "string",
        "Spark Settings State": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"/v1/workspaces/{workspace_id}/environments",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            pub = v.get("properties", {}).get("publishDetails", {})
            rows.append(
                {
                    "Environment Name": v.get("displayName"),
                    "Environment Id": v.get("id"),
                    "Description": v.get("description"),
                    "Publish State": pub.get("state"),
                    "Publish Target Version": pub.get("targetVersion"),
                    "Publish Start Time": pub.get("startTime"),
                    "Publish End Time": pub.get("endTime"),
                    "Spark Libraries State": pub.get("componentPublishInfo", {})
                    .get("sparkLibraries", {})
                    .get("state"),
                    "Spark Settings State": pub.get("componentPublishInfo", {})
                    .get("sparkSettings", {})
                    .get("state"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def delete_environment(environment: str | UUID, workspace: Optional[str | UUID] = None):
    """
    Deletes a Fabric environment.

    This is a wrapper function for the following API: `Items - Delete Environment <https://learn.microsoft.com/rest/api/fabric/environment/items/delete-environment>`_.

    Parameters
    ----------
    environment: str | uuid.UUID
        Name or ID of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=environment, type="Environment", workspace=workspace)


@log
def publish_environment(
    environment: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Publishes a Fabric environment.

    This is a wrapper function for the following API: `Spark Libraries - Publish Environment <https://learn.microsoft.com/rest/api/fabric/environment/spark-libraries/publish-environment>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    environment: str | uuid.UUID
        Name or ID of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = resolve_item_id(
        item=environment, type="Environment", workspace=workspace_id
    )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/environments/{item_id}/staging/publish",
        method="post",
        lro_return_status_code=True,
        status_codes=None,
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The '{environment}' environment within the '{workspace_name}' workspace has been published."
    )


@log
def cancel_publish_environment(
    environment: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Trigger an environment publish cancellation.

    This is a wrapper function for the following API: `Items - Cancel Publish Environment <https://learn.microsoft.com/rest/api/fabric/environment/items/cancel-publish-environment>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    environment: str | uuid.UUID
        Name or ID of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    item_id = resolve_item_id(
        item=environment, type="Environment", workspace=workspace_id
    )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/environments/{item_id}/staging/cancelPublish",
        method="post",
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} The publish of the '{environment}' environment within the '{workspace_name}' workspace has been cancelled."
    )



from typing import Optional, List
from uuid import UUID
from sempy_labs._helper_functions import (
    _create_spark_session,
    create_abfss_path,
    resolve_workspace_id,
    resolve_lakehouse_id,
    _get_delta_table,
)
from sempy._utils._log import log


@log
def _get_partitions(
    table_name: str,
    schema_name: Optional[str] = None,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
):

    workspace_id = resolve_workspace_id(workspace)
    lakehouse_id = resolve_lakehouse_id(lakehouse, workspace)
    path = create_abfss_path(lakehouse_id, workspace_id, table_name, schema_name)

    delta_table = _get_delta_table(path)
    details_df = delta_table.detail()

    return details_df.collect()[0].asDict()


@log
def is_partitioned(
    table: str,
    schema: Optional[str] = None,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
) -> bool:
    """
    Checks if a delta table is partitioned.

    Parameters
    ----------
    table : str
        The name of the delta table.
    schema : str, optional
        The schema of the table to check. If not provided, the default schema is used.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    bool
        True if the table is partitioned, False otherwise.
    """

    details = _get_partitions(
        table_name=table, schema_name=schema, lakehouse=lakehouse, workspace=workspace
    )
    return len(details["partitionColumns"]) > 0


@log
def list_partitioned_columns(
    table: str,
    schema: Optional[str] = None,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
) -> List[str]:
    """
    Lists the partitioned columns of a delta table.

    Parameters
    ----------
    table : str
        The name of the delta table.
    schema : str, optional
        The schema of the table to check. If not provided, the default schema is used.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    List[str]
        The list of partitioned columns.
    """

    details = _get_partitions(
        table_name=table, schema_name=schema, lakehouse=lakehouse, workspace=workspace
    )

    return details["partitionColumns"]


@log
def is_over_partitioned(
    table: str,
    schema: Optional[str] = None,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
    total_table_size_gb: int = 1000,
    average_partition_size_gb: int = 1,
) -> bool:
    """
    Checks if a delta table is over-partitioned.

    Parameters
    ----------
    table : str
        The name of the delta table.
    schema : str, optional
        The schema of the table to check. If not provided, the default schema is used.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    total_table_size_gb : int, default=1000
        Threshold for total table size in GB (default 1TB).
    average_partition_size_gb : int, default=1
        Threshold for average partition size in GB.

    Returns
    -------
    bool
        True if the table is over-partitioned, False otherwise.
    """

    workspace_id = resolve_workspace_id(workspace)
    lakehouse_id = resolve_lakehouse_id(lakehouse, workspace)
    path = create_abfss_path(lakehouse_id, workspace_id, table, schema)
    # Get DeltaTable details
    spark = _create_spark_session()
    details_df = spark.sql(f"DESCRIBE DETAIL delta.`{path}`")
    details = details_df.collect()[0].asDict()

    # Extract relevant fields
    size_bytes = details["sizeInBytes"]
    partition_cols = details["partitionColumns"]
    num_files = details["numFiles"]

    total_size_gb = size_bytes / (1024**3)

    # Only check if the table is partitioned
    if len(partition_cols) > 0 and num_files > 0:
        avg_partition_size_gb = total_size_gb / num_files

        if (
            total_size_gb < total_table_size_gb
            or avg_partition_size_gb < average_partition_size_gb
        ):
            return True

    return False



import pandas as pd
from typing import Optional
from uuid import UUID
from sempy._utils._log import log
import sempy_labs.mirrored_database as md


@log
def list_mirrored_databases(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the mirrored databases within a workspace.

    This is a wrapper function for the following API: `Items - List Mirrored Databases <https://learn.microsoft.com/rest/api/fabric/mirroredwarehouse/items/list-mirrored-databases>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the mirrored databases within a workspace.
    """

    return md.list_mirrored_databases(workspace=workspace)


@log
def create_mirrored_database(
    name: str, description: Optional[str] = None, workspace: Optional[str | UUID] = None
):
    """
    Creates a Fabric mirrored database.

    This is a wrapper function for the following API: `Items - Create Mirrored Database <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/items/create-mirrored-database>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name: str
        Name of the mirrored database.
    description : str, default=None
        A description of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    md.create_mirrored_database(name=name, description=description, workspace=workspace)


@log
def delete_mirrored_database(
    mirrored_database: str, workspace: Optional[str | UUID] = None
):
    """
    Deletes a mirrored database.

    This is a wrapper function for the following API: `Items - Delete Mirrored Database <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/items/delete-mirrored-database>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database: str
        Name of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    md.delete_mirrored_database(
        mirrored_database=mirrored_database, workspace=workspace
    )


@log
def get_mirroring_status(
    mirrored_database: str | UUID, workspace: Optional[str | UUID] = None
) -> str:
    """
    Get the status of the mirrored database.

    This is a wrapper function for the following API: `Mirroring - Get Mirroring Status <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/mirroring/get-mirroring-status>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database: str | uuid.UUID
        Name or ID of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    str
        The status of a mirrored database.
    """

    return md.get_mirroring_status(
        mirrored_database=mirrored_database, workspace=workspace
    )


@log
def get_tables_mirroring_status(
    mirrored_database: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Gets the mirroring status of the tables.

    This is a wrapper function for the following API: `Mirroring - Get Tables Mirroring Status <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/mirroring/get-tables-mirroring-status>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database: str | uuid.UUID
        Name or ID of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the mirroring status of the tables.
    """

    return md.get_tables_mirroring_status(
        mirrored_database=mirrored_database, workspace=workspace
    )


@log
def start_mirroring(
    mirrored_database: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Starts the mirroring for a database.

    This is a wrapper function for the following API: `Mirroring - Start Mirroring <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/mirroring/start-mirroring>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database: str | uuid.UUID
        Name or ID of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    md.start_mirroring(mirrored_database=mirrored_database, workspace=workspace)


@log
def stop_mirroring(
    mirrored_database: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Stops the mirroring for a database.

    This is a wrapper function for the following API: `Mirroring - Stop Mirroring <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/mirroring/stop-mirroring>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database: str | uuid.UUID
        Name or ID of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    md.stop_mirroring(mirrored_database=mirrored_database, workspace=workspace)


@log
def get_mirrored_database_definition(
    mirrored_database: str | UUID,
    workspace: Optional[str | UUID] = None,
    decode: bool = True,
) -> dict:
    """
    Obtains the mirrored database definition.

    This is a wrapper function for the following API: `Items - Get Mirrored Database Definition <https://learn.microsoft.com/rest/api/fabric/mirroreddatabase/items/get-mirrored-database-definition>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database : str | uuid.UUID
        The name or ID of the mirrored database.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    decode : bool, default=True
        If True, decodes the mirrored database definition file into .json format.
        If False, obtains the mirrored database definition file in base64 format.

    Returns
    -------
    dict
        The mirrored database definition.
    """

    return md.get_mirrored_database_definition(
        mirrored_database=mirrored_database,
        workspace=workspace,
        decode=decode,
    )


@log
def update_mirrored_database_definition(
    mirrored_database: str | UUID,
    mirrored_database_content: dict,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates an existing notebook with a new definition.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    mirrored_database : str | uuid.UUID
        The name or ID of the mirrored database to be updated.
    mirrored_database_content : dict
        The mirrored database definition (not in Base64 format).
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    md.update_mirrored_database_definition(
        mirrored_database=mirrored_database,
        mirrored_database_content=mirrored_database_content,
        workspace=workspace,
    )



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_id,
    _base_api,
    _create_dataframe,
    resolve_item_id,
    _update_dataframe_datatypes,
)
from uuid import UUID
from sempy._utils._log import log


def _get_spark_compute(
    environment: str | UUID,
    workspace: Optional[str | UUID] = None,
    staging: bool = False,
) -> pd.DataFrame:

    columns = {
        "Instance Pool Name": "string",
        "Instance Pool Type": "string",
        "Instance Pool Id": "string",
        "Driver Cores": "int",
        "Driver Memory": "string",
        "Executor Cores": "int",
        "Executor Memory": "string",
        "Dynamic Executor Allocation Enabled": "bool",
        "Dynamic Executor Allocation Min Executors": "int",
        "Dynamic Executor Allocation Max Executors": "int",
        "Spark Properties": "string",
        "Runtime Version": "string",
    }

    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_item_id(
        item=environment, type="Environment", workspace=workspace_id
    )

    url = f"/v1/workspaces/{workspace_id}/environments/{item_id}/sparkCompute"
    if staging:
        url = (
            f"/v1/workspaces/{workspace_id}/environments/{item_id}/staging/sparkCompute"
        )

    response = _base_api(
        request=url,
        client="fabric_sp",
    )

    rows = []
    rows.append(
        {
            "Instance Pool Name": response.get("instancePool", {}).get("name"),
            "Instance Pool Type": response.get("instancePool", {}).get("type"),
            "Instance Pool Id": response.get("instancePool", {}).get("id"),
            "Driver Cores": response.get("driverCores"),
            "Driver Memory": response.get("driverMemory"),
            "Executor Cores": response.get("executorCores"),
            "Executor Memory": response.get("executorMemory"),
            "Dynamic Executor Allocation Enabled": response.get(
                "dynamicExecutorAllocation", {}
            ).get("enabled"),
            "Dynamic Executor Allocation Min Executors": response.get(
                "dynamicExecutorAllocation", {}
            ).get("minExecutors"),
            "Dynamic Executor Allocation Max Executors": response.get(
                "dynamicExecutorAllocation", {}
            ).get("maxExecutors"),
            "Spark Properties": response.get("sparkProperties"),
            "Runtime Version": response.get("runtimeVersion"),
        }
    )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(df, columns)

    return df


@log
def get_published_spark_compute(
    environment: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Gets the Spark compute of a published Fabric environment.

    This is a wrapper function for the following API: `Published - Get Spark Compute <https://learn.microsoft.com/rest/api/fabric/environment/published/get-spark-compute>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    environment: str | uuid.UUID
        Name or ID of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    return _get_spark_compute(environment, workspace, staging=False)


@log
def get_staging_spark_compute(
    environment: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Gets the Spark compute of a staging Fabric environment.

    This is a wrapper function for the following API: `Staging - Get Spark Compute <https://learn.microsoft.com/rest/api/fabric/environment/staging/get-spark-compute>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    environment: str | uuid.UUID
        Name or ID of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    return _get_spark_compute(environment, workspace, staging=True)


def _list_libraries(
    environment: str | UUID,
    workspace: Optional[str | UUID] = None,
    staging: bool = False,
) -> pd.DataFrame:

    columns = {
        "Library Name": "string",
        "Library Type": "string",
        "Library Version": "string",
    }

    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)
    item_id = resolve_item_id(
        item=environment, type="Environment", workspace=workspace_id
    )

    url = f"/v1/workspaces/{workspace_id}/environments/{item_id}/libraries"
    if staging:
        url = f"/v1/workspaces/{workspace_id}/environments/{item_id}/staging/libraries"

    responses = _base_api(
        request=url,
        client="fabric_sp",
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        for lib in r.get("libraries", []):
            rows.append(
                {
                    "Library Name": lib.get("name"),
                    "Library Type": lib.get("libraryType"),
                    "Library Version": lib.get("version"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def list_published_libraries(
    environment: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Gets the published libraries of a Fabric environment.

    This is a wrapper function for the following API: `Published - List Libraries <https://learn.microsoft.com/rest/api/fabric/environment/published/list-libraries>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    environment: str | uuid.UUID
        Name or ID of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    return _list_libraries(environment, workspace, staging=False)


@log
def list_staging_libraries(
    environment: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Gets the staging libraries of a Fabric environment.

    This is a wrapper function for the following API: `Staging - List Libraries <https://learn.microsoft.com/rest/api/fabric/environment/staging/list-libraries>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    environment: str | uuid.UUID
        Name or ID of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    return _list_libraries(environment, workspace, staging=True)



from ._items import (
    create_on_prem_connection,
    list_connection_role_assignments,
    list_connections,
    delete_connection,
    delete_connection_role_assignment,
    list_item_connections,
    create_vnet_connection,
    create_cloud_connection,
)

__all__ = [
    "create_on_prem_connection",
    "list_connection_role_assignments",
    "list_connections",
    "delete_connection",
    "delete_connection_role_assignment",
    "list_item_connections",
    "create_vnet_connection",
    "create_cloud_connection",
]



from uuid import UUID
from typing import Optional, Literal
import pyarrow.dataset as ds
from sempy_labs._helper_functions import (
    _mount,
    delete_item,
    _base_api,
    resolve_workspace_name_and_id,
    resolve_lakehouse_name_and_id,
)
from sempy._utils._log import log
import sempy_labs._icons as icons
import os
import json


@log
def is_v_ordered(
    table_name: str,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
    schema: Optional[str] = None,
) -> bool:
    """
    Checks if a delta table is v-ordered.

    Parameters
    ----------
    table_name : str
        The name of the table to check.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    schema : str, optional
        The schema of the table to check. If not provided, the default schema is used.

    Returns
    -------
    bool
        True if the table is v-ordered, False otherwise.
    """

    local_path = _mount(lakehouse=lakehouse, workspace=workspace)
    table_path = (
        f"{local_path}/Tables/{schema}/{table_name}"
        if schema
        else f"{local_path}/Tables/{table_name}"
    )
    ds_schema = ds.dataset(table_path).schema.metadata

    if ds_schema:
        return any(b"vorder" in key for key in ds_schema.keys())

    delta_log_path = os.path.join(table_path, "_delta_log")

    def read_vorder_tag(delta_log_path):
        json_files = sorted(
            [f for f in os.listdir(delta_log_path) if f.endswith(".json")], reverse=True
        )

        if not json_files:
            return False

        latest_file = os.path.join(delta_log_path, json_files[0])

        with open(latest_file, "r") as f:
            all_data = [
                json.loads(line) for line in f if line.strip()
            ]  # one dict per line
            for data in all_data:
                if "metaData" in data:
                    return (
                        data.get("metaData", {})
                        .get("configuration", {})
                        .get("delta.parquet.vorder.enabled", "false")
                        == "true"
                    )

            # If no metaData, fall back to commitInfo
            for data in all_data:
                if "commitInfo" in data:
                    tags = data["commitInfo"].get("tags", {})
                    return tags.get("VORDER", "false").lower() == "true"

        return False  # Default if not found

    return read_vorder_tag(delta_log_path)


@log
def delete_lakehouse(
    lakehouse: str | UUID, workspace: Optional[str | UUID] = None
) -> None:
    """
    Deletes a lakehouse.

    This is a wrapper function for the following API: `Items - Delete Lakehouse <https://learn.microsoft.com/rest/api/fabric/lakehouse/items/delete-lakehouse>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    lakehouse : str | uuid.UUID
        The name or ID of the lakehouse to delete.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=lakehouse, item_type="lakehouse", workspace=workspace)


@log
def update_lakehouse(
    name: Optional[str] = None,
    description: Optional[str] = None,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Updates a lakehouse.

    This is a wrapper function for the following API: `Items - Update Lakehouse <https://learn.microsoft.com/rest/api/fabric/lakehouse/items/update-lakehouse>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name: str, default=None
        The new name of the lakehouse.
        Defaults to None which does not update the name.
    description: str, default=None
        The new description of the lakehouse.
        Defaults to None which does not update the description.
    lakehouse : str | uuid.UUID, default=None
        The name or ID of the lakehouse to update.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    if not name and not description:
        raise ValueError(
            f"{icons.red_dot} Either name or description must be provided."
        )

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse, workspace_id
    )

    payload = {}
    if name:
        payload["displayName"] = name
    if description:
        payload["description"] = description

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/lakehouses/{lakehouse_id}",
        method="patch",
        client="fabric_sp",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The '{lakehouse_name}' lakehouse within the '{workspace_name}' workspace has been updated accordingly."
    )


@log
def load_table(
    table_name: str,
    file_path: str,
    mode: Literal["Overwrite", "Append"],
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Loads a table into a lakehouse. Currently only files are supported, not folders.

    This is a wrapper function for the following API: `Tables - Load Table <https://learn.microsoft.com/rest/api/fabric/lakehouse/tables/load-table>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    table_name : str
        The name of the table to load.
    file_path : str
        The path to the data to load.
    mode : typing.Literal["Overwrite", "Append"]
        The mode to use when loading the data.
        "Overwrite" will overwrite the existing data.
        "Append" will append the data to the existing data.
    lakehouse : str | uuid.UUID, default=None
        The name or ID of the lakehouse to load the table into.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse, workspace_id
    )

    file_extension = os.path.splitext(file_path)[1]

    payload = {
        "relativePath": file_path,
        "pathType": "File",
        "mode": mode,
        "formatOptions": {},
    }

    if file_extension == ".csv":
        payload["formatOptions"] = {"format": "Csv", "header": True, "delimiter": ","}
    elif file_extension == ".parquet":
        payload["formatOptions"] = {
            "format": "Parquet",
            "header": True,
        }
    # Solve for loading folders
    # elif file_extension == '':
    #    payload['pathType'] = "Folder"
    #    payload["recursive"] = recursive
    #    payload['formatOptions']
    else:
        raise NotImplementedError()

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/lakehouses/{lakehouse_id}/tables/{table_name}/load",
        client="fabric_sp",
        method="post",
        status_codes=202,
        lro_return_status_code=True,
    )

    print(
        f"{icons.green_dot} The '{table_name}' table has been loaded into the '{lakehouse_name}' lakehouse within the '{workspace_name}' workspace."
    )



import sempy_labs._icons as icons
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    generate_guid,
)
from uuid import UUID
from sempy._utils._log import log
import requests
import json
from sempy_labs.kql_dashboard._items import create_kql_dashboard
from sempy_labs.kql_database._items import list_kql_databases


@log
def create_workspace_monitoring_dashboard(
    name: str = "Fabric Workspace Monitoring Dashboard",
    workspace: Optional[str | UUID] = None,
):
    """
    Creates a workspace monitoring dashboard based on `this <https://github.com/microsoft/fabric-toolbox/tree/main/monitoring/workspace-monitoring-dashboards>`_ template.

    This function requires the workspace to have workspace monitoring enabled.

    Parameters
    ----------
    name : str, default="Fabric Workspace Monitoring Dashboard"
        The name of the workspace monitoring dashboard.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    db_name = "Monitoring KQL database"

    url = "https://raw.githubusercontent.com/microsoft/fabric-toolbox/refs/heads/main/monitoring/workspace-monitoring-dashboards/Fabric%20Workspace%20Monitoring%20Dashboard.json"
    response = requests.get(url)
    content = json.loads(response.content)

    # Resolve the cluster URI and database ID
    dfK = list_kql_databases(workspace=workspace)
    dfK_filt = dfK[dfK["KQL Database Name"] == db_name]
    if dfK_filt.empty:
        raise ValueError(
            f"{icons.red_dot} Workspace monitoring is not set up for the '{workspace_name}' workspace."
        )
    cluster_uri = dfK_filt["Query Service URI"].iloc[0]
    database_id = dfK_filt["KQL Database Id"].iloc[0]

    content["dataSources"] = [
        {
            "kind": "kusto-trident",
            "scopeId": "kusto-trident",
            "clusterUri": cluster_uri,
            "database": database_id,
            "name": db_name,
            "id": generate_guid(),
            "workspace": workspace_id,
        }
    ]

    create_kql_dashboard(name=name, content=content, workspace=workspace)



import os
import pandas as pd
import pyarrow.parquet as pq
from datetime import datetime
from sempy_labs._helper_functions import (
    _get_column_aggregate,
    resolve_lakehouse_name_and_id,
    save_as_delta_table,
    resolve_workspace_id,
    _read_delta_table,
    _get_delta_table,
    _mount,
    create_abfss_path,
    _pure_python_notebook,
)
from sempy_labs.directlake._guardrails import (
    get_sku_size,
    get_directlake_guardrails_for_sku,
)
from sempy_labs.lakehouse._lakehouse import lakehouse_attached
from typing import Optional
import sempy_labs._icons as icons
from sempy._utils._log import log
from uuid import UUID
from sempy_labs.lakehouse._schemas import list_tables


@log
def get_lakehouse_tables(
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
    extended: bool = False,
    count_rows: bool = False,
    export: bool = False,
    exclude_shortcuts: bool = False,
) -> pd.DataFrame:
    """
    Shows the tables of a lakehouse and their respective properties. Option to include additional properties relevant to Direct Lake guardrails.

    This function can be executed in either a PySpark or pure Python notebook.

    This is a wrapper function for the following API: `Tables - List Tables <https://learn.microsoft.com/rest/api/fabric/lakehouse/tables/list-tables>`_ plus extended capabilities.
    However, the above mentioned API does not support Lakehouse schemas (Preview) until it is in GA (General Availability). This version also supports schema
    enabled Lakehouses.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    extended : bool, default=False
        Obtains additional columns relevant to the size of each table.
    count_rows : bool, default=False
        Obtains a row count for each lakehouse table.
    export : bool, default=False
        Exports the resulting dataframe to a delta table in the lakehouse.
    exclude_shortcuts : bool, default=False
        If True, excludes shortcuts.

    Returns
    -------
    pandas.DataFrame
        Shows the tables/columns within a lakehouse and their properties.
    """

    workspace_id = resolve_workspace_id(workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse=lakehouse, workspace=workspace_id
    )

    df = list_tables(lakehouse=lakehouse, workspace=workspace)

    local_path = _mount(lakehouse=lakehouse_id, workspace=workspace_id)

    if extended:
        sku_value = get_sku_size(workspace_id)
        guardrail = get_directlake_guardrails_for_sku(sku_value)
        # Avoid mounting the lakehouse if is already mounted
        if not local_path:
            local_path = _mount(lakehouse=lakehouse_id, workspace=workspace_id)

        df["Files"], df["Row Groups"], df["Table Size"] = None, None, None
        if count_rows:
            df["Row Count"] = None

        for i, r in df.iterrows():
            schema_name = r["Schema Name"]
            table_name = r["Table Name"]
            if r["Type"] == "Managed" and r["Format"] == "delta":
                delta_table_path = (
                    create_abfss_path(
                        lakehouse_id, workspace_id, table_name, schema_name
                    )
                    .replace("//", "/")  # When schema_name = ""
                    .replace("abfss:/", "abfss://")  # Put back the // after abfss:
                )

                if _pure_python_notebook():
                    from deltalake import DeltaTable

                    delta_table = DeltaTable(delta_table_path)
                    latest_files = [
                        file["path"]
                        for file in delta_table.get_add_actions().to_pylist()
                    ]
                    size_in_bytes = 0
                    for f in latest_files:
                        local_file_path = os.path.join(
                            local_path, "Tables", schema_name, table_name, f
                        )

                        if os.path.exists(local_file_path):
                            size_in_bytes += os.path.getsize(local_file_path)
                    num_latest_files = len(latest_files)
                else:
                    delta_table = _get_delta_table(delta_table_path)

                    latest_files = _read_delta_table(delta_table_path).inputFiles()
                    table_df = delta_table.toDF()
                    table_details = delta_table.detail().collect()[0].asDict()
                    size_in_bytes = table_details.get("sizeInBytes", 0)
                    num_latest_files = table_details.get("numFiles", 0)

                table_path = os.path.join(local_path, "Tables", schema_name, table_name)

                file_paths = []
                for file in latest_files:
                    if _pure_python_notebook():
                        file_paths.append(file)
                    else:
                        # Append the <Partition folder>/<filename> or <filename>
                        find_table = file.find(table_name)
                        len_file = len(file)
                        len_table = len(table_name)
                        last_chars = len_file - (find_table + len_table + 1)
                        file_paths.append(file[-last_chars:])

                num_rowgroups = 0
                for filename in file_paths:
                    parquet_file_path = f"{table_path}/{filename}"
                    if os.path.exists(parquet_file_path):
                        parquet_file = pq.ParquetFile(parquet_file_path)
                        num_rowgroups += parquet_file.num_row_groups

                df.at[i, "Files"] = num_latest_files
                df.at[i, "Row Groups"] = num_rowgroups
                df.at[i, "Table Size"] = size_in_bytes

            if count_rows:
                if _pure_python_notebook():
                    row_count = delta_table.to_pyarrow_table().num_rows
                else:
                    row_count = table_df.count()
                df.at[i, "Row Count"] = row_count

            # Set "Schema Name" = "dbo" when it is ""
            df.loc[df["Schema Name"] == "", "Schema Name"] = "dbo"

    if extended:
        intColumns = ["Files", "Row Groups", "Table Size"]
        df[intColumns] = df[intColumns].astype(int)

        col_name = guardrail.columns[0]
        df["SKU"] = guardrail[col_name].iloc[0]
        df["Parquet File Guardrail"] = guardrail["Parquet files per table"].iloc[0]
        df["Row Group Guardrail"] = guardrail["Row groups per table"].iloc[0]
        df["Row Count Guardrail"] = (
            guardrail["Rows per table (millions)"].iloc[0] * 1000000
        )

        df["Parquet File Guardrail Hit"] = df["Files"] > df["Parquet File Guardrail"]
        df["Row Group Guardrail Hit"] = df["Row Groups"] > df["Row Group Guardrail"]
    if count_rows:
        df["Row Count"] = df["Row Count"].astype(int)
        df["Row Count Guardrail Hit"] = df["Row Count"] > df["Row Count Guardrail"]

    if exclude_shortcuts:
        from sempy_labs.lakehouse._shortcuts import list_shortcuts

        # Exclude shortcuts
        shortcuts = (
            list_shortcuts(lakehouse=lakehouse, workspace=workspace)
            .query("`Shortcut Path`.str.startswith('/Tables')", engine="python")
            .assign(
                FullPath=lambda df: df["Shortcut Path"].str.rstrip("/")
                + "/"
                + df["Shortcut Name"]
            )["FullPath"]
            .tolist()
        )

        df["FullPath"] = df.apply(
            lambda x: (
                f"/Tables/{x['Table Name']}"
                if pd.isna(x["Schema Name"]) or x["Schema Name"] == ""
                else f"/Tables/{x['Schema Name']}/{x['Table Name']}"
            ),
            axis=1,
        )

        df = df[~df["FullPath"].isin(shortcuts)].reset_index(drop=True)

    if export:
        if not lakehouse_attached():
            raise ValueError(
                f"{icons.red_dot} In order to save the dataframe, a lakehouse must be attached to the notebook. Please attach a lakehouse to this notebook."
            )

        lake_table_name = "lakehouse_table_details"
        df_filt = df[df["Table Name"] == lake_table_name]

        if df_filt.empty:
            run_id = 1
        else:
            max_run_id = _get_column_aggregate(table_name=lake_table_name)
            run_id = max_run_id + 1

        export_df = df.copy()

        cols = [
            "Files",
            "Row Groups",
            "Row Count",
            "Table Size",
            "SKU",
            "Parquet File Guardrail",
            "Row Group Guardrail",
            "Row Count Guardrail",
            "Parquet File Guardrail Hit",
            "Row Group Guardrail Hit",
            "Row Count Guardrail Hit",
        ]

        for c in cols:
            if c not in export_df:
                if c in [
                    "Files",
                    "Row Groups",
                    "Row Count",
                    "Table Size",
                    "Parquet File Guardrail",
                    "Row Group Guardrail",
                    "Row Count Guardrail",
                ]:
                    export_df[c] = 0
                    export_df[c] = export_df[c].astype(int)
                elif c in ["SKU"]:
                    export_df[c] = None
                    export_df[c] = export_df[c].astype(str)
                elif c in [
                    "Parquet File Guardrail Hit",
                    "Row Group Guardrail Hit",
                    "Row Count Guardrail Hit",
                ]:
                    export_df[c] = False
                    export_df[c] = export_df[c].astype(bool)

        print(
            f"{icons.in_progress} Saving Lakehouse table properties to the '{lake_table_name}' table in the lakehouse...\n"
        )
        export_df["Timestamp"] = datetime.now()
        export_df["RunId"] = run_id

        save_as_delta_table(
            dataframe=export_df, delta_table_name=lake_table_name, write_mode="append"
        )

    return df



import pandas as pd
from uuid import UUID
from sempy_labs._helper_functions import (
    _is_valid_uuid,
    _base_api,
    _update_dataframe_datatypes,
    _create_dataframe,
)


_pbi_url_prefix = None
_theme_url_prefix = "metadata/v202409/organization/themes"


def init_pbi_url_prefix():
    global _pbi_url_prefix
    if _pbi_url_prefix:
        return

    response = _base_api("/v1.0/myorg/capacities")
    _pbi_url_prefix = response.json().get("@odata.context").split("/v1.0/myorg")[0]


def list_org_themes() -> pd.DataFrame:
    """
    Lists all `organizational themes <https://www.linkedin.com/pulse/organizational-themes-preview-pbicorevisuals-j7jxe/>`_ in Power BI. Note that this uses an internal API and may break at any time.

    Returns
    -------
    pandas.DataFrame
        A dataframe containing the details of the organizational themes.
    """

    init_pbi_url_prefix()

    columns = {
        "Theme Id": "str",
        "Theme Name": "str",
        "Description": "str",
        "Enabled In Theme Gallery": "bool",
        "Update Time": "datetime",
        "For Copilot": "bool",
        "Theme Json": "str",
    }

    df = _create_dataframe(columns=columns)

    response = _base_api(request=f"{_pbi_url_prefix}/{_theme_url_prefix}")

    result = response.json().get("orgThemes")
    if result:
        df = pd.json_normalize(result)
        df.rename(
            columns={
                "objectId": "Theme Id",
                "displayName": "Theme Name",
                "description": "Description",
                "enabledInThemeGallery": "Enabled In Theme Gallery",
                "updateTime": "Update Time",
                "forCopilot": "For Copilot",
                "themeJson": "Theme Json",
            },
            inplace=True,
        )
        _update_dataframe_datatypes(df, columns)

    return df


def resolve_theme_id(theme: str | UUID):
    """
    Resolves the ID of an organization theme by its name or ID.

    Parameters
    ----------
    theme : str | uuid.UUID
        The name or ID of the organization theme.

    Returns
    -------
    uuid.UUID
        The ID of the organization theme.
    """

    if _is_valid_uuid(theme):
        return theme
    else:
        df = list_org_themes()
        df_filt = df[df["Theme Name"] == theme]
        if df_filt.empty:
            raise ValueError("Invalid theme")
        else:
            return df_filt["Theme Id"].iloc[0]


def get_org_theme_json(theme: str | UUID) -> dict:
    """
    Retrieves the JSON representation of an organization theme by its name or ID.

    Parameters
    ----------
    theme : str | uuid.UUID
        The name or ID of the organization theme.

    Returns
    -------
    dict
        The JSON representation of the organization theme.
    """

    init_pbi_url_prefix()

    theme_id = resolve_theme_id(theme)
    response = _base_api(request=f"{_pbi_url_prefix}/{_theme_url_prefix}/{theme_id}")
    return response.json().get("themeJson", {})


def delete_org_theme(theme: str | UUID) -> None:
    """
    Deletes an organization theme by its name or ID.
    """
    init_pbi_url_prefix()

    theme_id = resolve_theme_id(theme)
    _base_api(
        request=f"{_pbi_url_prefix}/{_theme_url_prefix}/{theme_id}",
        method="delete",
        status_codes=204,
    )



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_item_name_and_id,
    resolve_workspace_id,
    _base_api,
    _create_dataframe,
    delete_item,
    resolve_item_id,
    resolve_workspace_name_and_id,
)
from uuid import UUID
from sempy._utils._log import log
import sempy_labs._icons as icons


@log
def list_apache_airflow_job_files(
    apache_airflow_job: str | UUID,
    workspace: Optional[str | UUID] = None,
    root_path: Optional[str] = None,
) -> pd.DataFrame:
    """
    Shows a list of Apache Airflow job files from the specified Apache Airflow job.

    This is a wrapper function for the following API: `Files - List Apache Airflow Job Files <https://learn.microsoft.com/rest/api/fabric/apacheairflowjob/files/list-apache-airflow-job-files>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    apache_airflow_job : str | uuid.UUID
        The Apache Airflow job name or ID.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    root_path : str, default=None
        The folder path to list. If not provided, the root directory is used.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of Apache Airflow job files from the specified Apache Airflow job.
    """

    columns = {
        "File Path": "string",
        "Size In Bytes": "int",
    }

    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    apache_airflow_job_id = resolve_item_id(
        item=apache_airflow_job,
        type="ApacheAirflowJob",
        workspace=workspace_id,
    )

    url = f"/v1/workspaces/{workspace_id}/ApacheAirflowJobs/{apache_airflow_job_id}/files?beta=True"
    if root_path:
        url += f"&rootPath={root_path}"
    responses = _base_api(
        request=url,
        client="fabric_sp",
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "File Path": v.get("filePath"),
                    "Size In Bytes": v.get("sizeInBytes"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def delete_apache_airflow_job_file(
    apache_airflow_job: str | UUID,
    file_path: str,
    workspace: Optional[str | UUID] = None,
):
    """
    Deletes a file from a Fabric Apache Airflow Job.

    This is a wrapper function for the following API: `Files - Delete Apache Airflow Job File <https://learn.microsoft.com/rest/api/fabric/apacheairflowjob/files/delete-apache-airflow-job-file>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    apache_airflow_job : str | uuid.UUID
        The Apache Airflow job name or ID.
    file_path : str
        The file path relative to the Apache Airflow job root. It must begin with either 'dags/' or 'plugins/' (for example, dags/example_dag.py).
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(
        item=apache_airflow_job,
        type="ApacheAirflowJob",
        workspace=workspace_id,
    )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/ApacheAirflowJobs/{item_id}/files/{file_path}?beta=True",
        client="fabric_sp",
        method="delete",
    )

    print(
        f"{icons.green_dot} The '{file_path}' file has been deleted from the '{item_name}' Apache Airflow Job within the '{workspace_name}' workspace."
    )


@log
def create_or_update_apache_airflow_job_file(
    apache_airflow_job: str | UUID,
    file_path: str,
    file_content: bytes,
    workspace: Optional[str | UUID] = None,
):
    """
    Creates or updates an Apache Airflow job file.

    This is a wrapper function for the following API: `Files - Create Or Update Apache Airflow Job File <https://learn.microsoft.com/rest/api/fabric/apacheairflowjob/files/create-or-update-apache-airflow-job-file>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    apache_airflow_job : str | uuid.UUID
        The Apache Airflow job name or ID.
    file_path : str
        The file path relative to the Apache Airflow job root. It must begin with either 'dags/' or 'plugins/' (for example, dags/example_dag.py).
    file_content : bytes
        The file content. Text files must be UTF-8 encoded.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(
        item=apache_airflow_job,
        type="ApacheAirflowJob",
        workspace=workspace_id,
    )

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/ApacheAirflowJobs/{item_id}/files/{file_path}?beta=True",
        client="fabric_sp",
        method="put",
        payload=file_content,
    )

    print(
        f"{icons.green_dot} The '{file_path}' file has been created/updated in the '{item_name}' Apache Airflow Job within the '{workspace_name}' workspace."
    )



import pandas as pd
from sempy_labs._helper_functions import (
    _base_api,
    _create_dataframe,
)
from sempy._utils._log import log


@log
def list_domains() -> pd.DataFrame:
    """
    Returns a list of all the tenant's domains.

    This is a wrapper function for the following API: `Domains - List Domains <https://learn.microsoft.com/rest/api/fabric/core/domains/list-domains>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of the domains.
    """

    columns = {
        "Domain Id": "string",
        "Domain Name": "string",
        "Description": "string",
        "Parent Domain Id": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request="/v1/domains", client="fabric_sp", uses_pagination=True
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "Domain Id": v.get("id"),
                    "Domain Name": v.get("displayName"),
                    "Description": v.get("description"),
                    "Parent Domain Id": v.get("parentDomainId"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df



from ._items import (
    list_gateway_members,
    list_gateway_role_assigments,
    list_gateways,
    delete_gateway,
    delete_gateway_member,
    delete_gateway_role_assignment,
    create_vnet_gateway,
    update_vnet_gateway,
    update_on_premises_gateway,
    bind_semantic_model_to_gateway,
)

__all__ = [
    "list_gateway_members",
    "list_gateway_role_assigments",
    "list_gateways",
    "delete_gateway",
    "delete_gateway_member",
    "delete_gateway_role_assignment",
    "create_vnet_gateway",
    "update_vnet_gateway",
    "update_on_premises_gateway",
    "bind_semantic_model_to_gateway",
]



from sempy_labs._helper_functions import (
    _update_dataframe_datatypes,
    _base_api,
    _create_dataframe,
)
from sempy._utils._log import log
import pandas as pd
from uuid import UUID
from sempy_labs.admin._capacities import _resolve_capacity_name_and_id
import sempy_labs._icons as icons
from typing import Optional, List


@log
def list_tenant_settings() -> pd.DataFrame:
    """
    Lists all tenant settings.

    This is a wrapper function for the following API: `Tenants - List Tenant Settings <https://learn.microsoft.com/rest/api/fabric/admin/tenants/list-tenant-settings>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the tenant settings.
    """

    columns = {
        "Setting Name": "string",
        "Title": "string",
        "Enabled": "bool",
        "Can Specify Security Groups": "bool",
        "Tenant Setting Group": "string",
        "Enabled Security Groups": "list",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(request="/v1/admin/tenantsettings", client="fabric_sp")

    rows = []
    for i in response.json().get("value", []):
        rows.append(
            {
                "Setting Name": i.get("settingName"),
                "Title": i.get("title"),
                "Enabled": i.get("enabled"),
                "Can Specify Security Groups": i.get("canSpecifySecurityGroups"),
                "Tenant Setting Group": i.get("tenantSettingGroup"),
                "Enabled Security Groups": [i.get("enabledSecurityGroups", [])],
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def list_capacity_tenant_settings_overrides(
    capacity: Optional[str | UUID] = None,
    return_dataframe: bool = True,
) -> pd.DataFrame | dict:
    """
    Returns list of tenant setting overrides that override at the capacities.

    This is a wrapper function for the following API: `Tenants - List Capacities Tenant Settings Overrides <https://learn.microsoft.com/rest/api/fabric/admin/tenants/list-capacities-tenant-settings-overrides>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        The capacity name or ID.
        Defaults to None which resolves to showing all capacities.
    return_dataframe : bool, default=True
        If True, returns a dataframe. If False, returns a dictionary.

    Returns
    -------
    pandas.DataFrame | dict
        A pandas dataframe showing a list of tenant setting overrides that override at the capacities.
    """

    columns = {
        "Capacity Id": "string",
        "Setting Name": "string",
        "Setting Title": "string",
        "Setting Enabled": "bool",
        "Can Specify Security Groups": "bool",
        "Enabled Security Groups": "list",
        "Tenant Setting Group": "string",
        "Tenant Setting Properties": "list",
        "Delegate to Workspace": "bool",
        "Delegated From": "string",
    }

    if capacity is None:
        url = "/v1/admin/capacities/delegatedTenantSettingOverrides"
    else:
        (_, capacity_id) = _resolve_capacity_name_and_id(capacity=capacity)
        url = f"/v1/admin/capacities/{capacity_id}/delegatedTenantSettingOverrides"
    responses = _base_api(
        request=url,
        client="fabric_sp",
        uses_pagination=True,
    )

    def create_new_data(setting, capacity_id=None):
        return {
            "Capacity Id": capacity_id or setting.get("id"),
            "Setting Name": setting.get("settingName"),
            "Setting Title": setting.get("title"),
            "Setting Enabled": setting.get("enabled"),
            "Can Specify Security Groups": setting.get("canSpecifySecurityGroups"),
            "Enabled Security Groups": setting.get("enabledSecurityGroups", []),
            "Tenant Setting Group": setting.get("tenantSettingGroup"),
            "Tenant Setting Properties": setting.get("properties", []),
            "Delegate to Workspace": setting.get("delegateToWorkspace"),
            "Delegated From": setting.get("delegatedFrom"),
        }

    def process_responses(responses, capacity_id=None, return_dataframe=False):
        data = []
        df = _create_dataframe(columns=columns)

        for r in responses:
            if capacity_id is None:
                # If capacity_id is None, we access 'Overrides' -> 'tenantSettings'
                for override in r.get("overrides", []):
                    capacity_id = override.get("id")
                    tenant_settings = override.get("tenantSettings", [])
                    for setting in tenant_settings:
                        data.append(create_new_data(setting, capacity_id))
            else:
                # If capacity_id is provided, we access 'value' directly for tenantSettings
                for setting in r.get("value", []):
                    data.append(
                        create_new_data(setting, capacity_id)
                    )  # Use provided capacity_id

        if return_dataframe:
            if data:
                df = pd.DataFrame(data)
                _update_dataframe_datatypes(dataframe=df, column_map=columns)
            return df
        else:
            key = "overrides" if capacity_id is None else "value"
            continuation_uri = r.get("continuationUri", "")
            continuation_token = r.get("continuationToken", "")

            return {
                key: data,
                "continuationUri": continuation_uri,
                "continuationToken": continuation_token,
            }

    # Main logic
    if capacity is None:
        return (
            process_responses(responses, return_dataframe=True)
            if return_dataframe
            else process_responses(responses)
        )
    else:
        return (
            process_responses(responses, capacity_id=capacity_id, return_dataframe=True)
            if return_dataframe
            else process_responses(responses, capacity_id=capacity_id)
        )


@log
def list_capacities_delegated_tenant_settings(
    return_dataframe: bool = True,
) -> pd.DataFrame | dict:
    """
    Returns list of tenant setting overrides that override at the capacities.

    NOTE: This function is to be deprecated. Please use the `list_capacity_tenant_settings_overrides` function instead.

    This is a wrapper function for the following API: `Tenants - List Capacities Tenant Settings Overrides <https://learn.microsoft.com/rest/api/fabric/admin/tenants/list-capacities-tenant-settings-overrides>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    return_dataframe : bool, default=True
        If True, returns a dataframe. If False, returns a dictionary.

    Returns
    -------
    pandas.DataFrame | dict
        A pandas dataframe showing a list of tenant setting overrides that override at the capacities.
    """

    list_capacity_tenant_settings_overrides(return_dataframe=return_dataframe)


@log
def delete_capacity_tenant_setting_override(capacity: str | UUID, tenant_setting: str):
    """
    Remove given tenant setting override for given capacity Id.

    This is a wrapper function for the following API: `Tenants - Delete Capacity Tenant Setting Override <https://learn.microsoft.com/rest/api/fabric/admin/tenants/delete-capacity-tenant-setting-override>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity : str | uuid.UUID
        The capacity name or ID.
    tenant_setting : str
        The tenant setting name. Example: "TenantSettingForCapacityDelegatedSwitch"
    """

    (capacity_name, capacity_id) = _resolve_capacity_name_and_id(capacity=capacity)

    _base_api(
        request=f"/v1/admin/capacities/{capacity_id}/delegatedTenantSettingOverrides/{tenant_setting}",
        client="fabric_sp",
        method="delete",
    )

    print(
        f"{icons.green_dot} The '{tenant_setting}' tenant setting has been removed from the '{capacity_name}' capacity."
    )


@log
def update_tenant_setting(
    tenant_setting: str,
    enabled: bool,
    delegate_to_capacity: Optional[bool] = None,
    delegate_to_domain: Optional[bool] = None,
    delegate_to_workspace: Optional[bool] = None,
    enabled_security_groups: Optional[List[dict]] = None,
    excluded_security_groups: Optional[List[dict]] = None,
    properties: Optional[List[dict]] = None,
):
    """
    Update a given tenant setting.

    This is a wrapper function for the following API: `Tenants - Update Tenant Setting <https://learn.microsoft.com/rest/api/fabric/admin/tenants/update-tenant-setting>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    tenant_setting : str
        The tenant setting name. Example: "TenantSettingForCapacityDelegatedSwitch"
    enabled : bool
        The status of the tenant setting. False - Disabled, True - Enabled.
    delegate_to_capacity : bool, default=None
        Indicates whether the tenant setting can be delegated to a capacity admin. False - Capacity admin cannot override the tenant setting. True - Capacity admin can override the tenant setting.
    delegate_to_domain : bool, default=None
        Indicates whether the tenant setting can be delegated to a domain admin. False - Domain admin cannot override the tenant setting. True - Domain admin can override the tenant setting.
    delegate_to_workspace : bool, default=None
        Indicates whether the tenant setting can be delegated to a workspace admin. False - Workspace admin cannot override the tenant setting. True - Workspace admin can override the tenant setting.
    enabled_security_groups : List[dict], default=None
        A list of enabled security groups. Example:
        [
            {
            "graphId": "f51b705f-a409-4d40-9197-c5d5f349e2f0",
            "name": "TestComputeCdsa"
            }
        ]
    excluded_security_groups : List[dict], default=None
        A list of excluded security groups. Example:
        [
            {
            "graphId": "f51b705f-a409-4d40-9197-c5d5f349e2f0",
            "name": "TestComputeCdsa"
            }
        ]
    properties : List[dict], default=None
        Tenant setting properties. Example:
        [
            {
            "name": "CreateP2w",
            "value": "true",
            "type": "Boolean"
            }
        ]
    """

    payload = {"enabled": enabled}

    if delegate_to_capacity is not None:
        payload["delegateToCapacity"] = delegate_to_capacity
    if delegate_to_domain is not None:
        payload["delegateToDomain"] = delegate_to_domain
    if delegate_to_workspace is not None:
        payload["delegateToWorkspace"] = delegate_to_workspace
    if enabled_security_groups is not None:
        payload["enabledSecurityGroups"] = enabled_security_groups
    if excluded_security_groups is not None:
        payload["excludedSecurityGroups"] = excluded_security_groups
    if properties is not None:
        payload["properties"] = properties

    _base_api(
        request=f"/v1/admin/tenantsettings/{tenant_setting}/update",
        client="fabric_sp",
        method="post",
        payload=payload,
    )

    print(f"{icons.green_dot} The '{tenant_setting}' tenant setting has been updated.")


@log
def update_capacity_tenant_setting_override(
    capacity: str | UUID,
    tenant_setting: str,
    enabled: bool,
    delegate_to_workspace: Optional[bool] = None,
    enabled_security_groups: Optional[List[dict]] = None,
    excluded_security_groups: Optional[List[dict]] = None,
):
    """
    Update given tenant setting override for given capacity.

    This is a wrapper function for the following API: `Tenants - Update Capacity Tenant Setting Override <https://learn.microsoft.com/en-us/rest/api/fabric/admin/tenants/update-capacity-tenant-setting-override>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity : str | uuid.UUID
        The capacity name or ID.
    tenant_setting : str
        The tenant setting name. Example: "TenantSettingForCapacityDelegatedSwitch"
    enabled : bool
        The status of the tenant setting. False - Disabled, True - Enabled.
    delegate_to_workspace : bool, default=None
        Indicates whether the tenant setting can be delegated to a workspace admin. False - Workspace admin cannot override the tenant setting. True - Workspace admin can override the tenant setting.
    enabled_security_groups : List[dict], default=None
        A list of enabled security groups. Example:
        [
            {
            "graphId": "f51b705f-a409-4d40-9197-c5d5f349e2f0",
            "name": "TestComputeCdsa"
            }
        ]
    excluded_security_groups : List[dict], default=None
        A list of excluded security groups. Example:
        [
            {
            "graphId": "f51b705f-a409-4d40-9197-c5d5f349e2f0",
            "name": "TestComputeCdsa"
            }
        ]
    """

    (capacity_name, capacity_id) = _resolve_capacity_name_and_id(capacity=capacity)

    payload = {"enabled": enabled}

    if delegate_to_workspace is not None:
        payload["delegateToWorkspace"] = delegate_to_workspace
    if enabled_security_groups is not None:
        payload["enabledSecurityGroups"] = enabled_security_groups
    if excluded_security_groups is not None:
        payload["excludedSecurityGroups"] = excluded_security_groups

    _base_api(
        request=f"/v1/admin/capacities/{capacity_id}/delegatedTenantSettingOverrides/{tenant_setting}/update",
        client="fabric_sp",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The '{tenant_setting}' tenant setting for the '{capacity_name}' capacity has been updated."
    )


@log
def list_workspaces_tenant_settings_overrides() -> pd.DataFrame:
    """
    Shows a list of workspace delegation setting overrides. In order to run this function, you must enable the workspace's delegated OneLake settings. To do this, navigate to the workspace, Workspace Settings -> Delegated Settings -> OneLake settings -> Set to 'On'.

    This is a wrapper function for the following API: `Tenants - List Workspaces Tenant Settings Overrides <https://learn.microsoft.com/rest/api/fabric/admin/tenants/list-workspaces-tenant-settings-overrides>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of workspace delegation setting overrides.
    """

    columns = {
        "Workspace Id": "string",
        "Setting Name": "string",
        "Title": "string",
        "Enabled": "bool",
        "Can Specify Security Groups": "bool",
        "Enabled Security Groups": "list",
        "Tenant Setting Group": "string",
        "Delegated From": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request="/v1/admin/workspaces/delegatedTenantSettingOverrides",
        client="fabric_sp",
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            workspace_id = v.get("id")
            for setting in v.get("tenantSettings", []):
                rows.append(
                    {
                        "Workspace Id": workspace_id,
                        "Setting Name": setting.get("settingName"),
                        "Title": setting.get("title"),
                        "Enabled": setting.get("enabled"),
                        "Can Specify Security Groups": setting.get(
                            "canSpecifySecurityGroups"
                        ),
                        "Enabled Security Groups": [
                            setting.get("enabledSecurityGroups", [])
                        ],
                        "Tenant Setting Group": setting.get("tenantSettingGroup"),
                        "Delegated From": setting.get("delegatedFrom"),
                    }
                )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def list_domain_tenant_settings_overrides() -> pd.DataFrame:
    """
    Shows a list of domain delegation setting overrides.

    This is a wrapper function for the following API: `Tenants - List Domains Tenant Settings Overrides <https://learn.microsoft.com/rest/api/fabric/admin/tenants/list-domains-tenant-settings-overrides>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of domain delegation setting overrides.
    """

    columns = {
        "Domain Id": "string",
        "Setting Name": "string",
        "Title": "string",
        "Enabled": "bool",
        "Can Specify Security Groups": "bool",
        "Enabled Security Groups": "list",
        "Tenant Setting Group": "string",
        "Delegated To Workspace": "bool",
        "Delegated From": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request="/v1/admin/domains/delegatedTenantSettingOverrides",
        client="fabric_sp",
        uses_pagination=True,
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            domain_id = v.get("id")
            for setting in v.get("tenantSettings", []):
                rows.append(
                    {
                        "Domain Id": domain_id,
                        "Setting Name": setting.get("settingName"),
                        "Title": setting.get("title"),
                        "Enabled": setting.get("enabled"),
                        "Can Specify Security Groups": setting.get(
                            "canSpecifySecurityGroups"
                        ),
                        "Enabled Security Groups": [
                            setting.get("enabledSecurityGroups", [])
                        ],
                        "Tenant Setting Group": setting.get("tenantSettingGroup"),
                        "Delegated To Workspace": setting.get("delegateToWorkspace"),
                        "Delegated From": setting.get("delegatedFrom"),
                    }
                )
    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))
        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df



from typing import Optional, List, Union, Tuple
from uuid import UUID
import sempy_labs._icons as icons
from sempy_labs._helper_functions import (
    _is_valid_uuid,
    _build_url,
    _base_api,
    _create_dataframe,
)
from sempy._utils._log import log
import numpy as np
import pandas as pd
from dateutil.parser import parse as dtparser


@log
def list_workspaces(
    capacity: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
    workspace_state: Optional[str] = None,
    workspace_type: Optional[str] = None,
    **kwargs,
) -> pd.DataFrame:
    """
    Lists workspaces for the organization. This function is the admin version of list_workspaces.

    This is a wrapper function for the following API: `Workspaces - List Workspaces - REST API (Admin) <https://learn.microsoft.com/en-us/rest/api/fabric/admin/workspaces/list-workspaces>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    capacity : str | uuid.UUID, default=None
        Returns only the workspaces in the specified Capacity.
    workspace : str | uuid.UUID, default=None
        Returns the workspace with the specific name.
    workspace_state : str, default=None
        Return only the workspace with the requested state. You can find the possible states in `Workspace States <https://learn.microsoft.com/en-us/rest/api/fabric/admin/workspaces/list-workspaces?tabs=HTTP#workspacestate>`_.
    workspace_type : str, default=None
        Return only the workspace of the specific type. You can find the possible types in `Workspace Types <https://learn.microsoft.com/en-us/rest/api/fabric/admin/workspaces/list-workspaces?tabs=HTTP#workspacetype>`_.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of workspaces for the organization.
    """

    from sempy_labs.admin._capacities import _resolve_capacity_name_and_id

    if "filter" in kwargs:
        print(
            "The 'filter' parameter has been deprecated. Please remove this parameter from the function going forward."
        )
        del kwargs["filter"]

    if "top" in kwargs:
        print(
            "The 'top' parameter has been deprecated. Please remove this parameter from the function going forward."
        )
        del kwargs["top"]

    if "skip" in kwargs:
        print(
            "The 'skip' parameter has been deprecated. Please remove this parameter from the function going forward."
        )
        del kwargs["skip"]

    columns = {
        "Id": "string",
        "Name": "string",
        "State": "string",
        "Type": "string",
        "Capacity Id": "string",
    }
    df = _create_dataframe(columns=columns)

    url = "/v1/admin/workspaces"
    params = {}

    if capacity is not None:
        params["capacityId"] = _resolve_capacity_name_and_id(capacity)[1]

    if workspace is not None and not _is_valid_uuid(workspace):
        params["name"] = workspace

    if workspace_state is not None:
        params["state"] = workspace_state

    if workspace_type is not None:
        params["type"] = workspace_type

    url = _build_url(url, params)

    responses = _base_api(request=url, client="fabric_sp", uses_pagination=True)
    workspaces = []

    for r in responses:
        workspaces = workspaces + r.get("workspaces", [])

    if len(workspaces) > 0:
        df = pd.DataFrame(workspaces)
        df.rename(
            columns={
                "id": "Id",
                "name": "Name",
                "state": "State",
                "type": "Type",
                "capacityId": "Capacity Id",
            },
            inplace=True,
        )

        df["Capacity Id"] = df["Capacity Id"].str.lower()

        if workspace is not None and _is_valid_uuid(workspace):
            df = df[df["Id"] == workspace.lower()]

    return df


@log
def assign_workspaces_to_capacity(
    source_capacity: Optional[str | UUID] = None,
    target_capacity: Optional[str | UUID] = None,
    workspace: Optional[str | List[str] | UUID | List[UUID]] = None,
):
    """
    Assigns a workspace to a capacity. This function is the admin version.

    This is a wrapper function for the following API: `Admin - Capacities AssignWorkspacesToCapacity <https://learn.microsoft.com/rest/api/power-bi/admin/capacities-assign-workspaces-to-capacity>`_.

    Parameters
    ----------
    source_capacity : str | uuid.UUID, default=None
        The name of the source capacity. If the Workspace is not specified, this is parameter mandatory.
    target_capacity : str | uuid.UUID, default=None
        The name of the target capacity.
    workspace : str | List[str] | uuid.UUID | List[uuid.UUID], default=None
        The name or ID of the workspace(s).
        Defaults to None which resolves to migrating all workspaces within the source capacity to the target capacity.
    """
    from sempy_labs.admin._capacities import _resolve_capacity_name_and_id

    if target_capacity is None:
        raise ValueError(
            f"{icons.red_dot} The parameter 'target_capacity' is mandatory."
        )

    if source_capacity is None and workspace is None:
        raise ValueError(
            f"{icons.red_dot} The parameters 'source_capacity' or 'workspace' needs to be specified."
        )

    if workspace is None:
        source_capacity_id = _resolve_capacity_name_and_id(source_capacity)[1]
        dfW = list_workspaces(capacity=source_capacity_id)
        workspaces = dfW["Id"].tolist()
    else:
        if isinstance(workspace, str) or isinstance(workspace, UUID):
            workspace = [workspace]
        if source_capacity is None:
            dfW = list_workspaces()
        else:
            source_capacity_id = _resolve_capacity_name_and_id(source_capacity)[1]
            dfW = list_workspaces(capacity=source_capacity_id)

        # Extract names and IDs that are mapped in dfW
        workspaces_names = dfW[dfW["Name"].isin(workspace)]["Name"].tolist()
        workspaces_ids = dfW[dfW["Id"].isin(workspace)]["Id"].tolist()

        # Combine IDs into the final workspaces list
        workspaces = workspaces_ids + dfW[dfW["Name"].isin(workspace)]["Id"].tolist()

        # Identify unmapped workspaces
        unmapped_workspaces = [
            item
            for item in workspace
            if item not in workspaces_names and item not in workspaces_ids
        ]

        if len(workspace) != len(workspaces):
            raise ValueError(
                f"{icons.red_dot} The following workspaces are invalid or not found in source capacity: {unmapped_workspaces}."
            )

    target_capacity_id = _resolve_capacity_name_and_id(target_capacity)[1]

    workspaces = np.array(workspaces)
    batch_size = 999
    for i in range(0, len(workspaces), batch_size):
        batch = workspaces[i : i + batch_size].tolist()
        payload = {
            "capacityMigrationAssignments": [
                {
                    "targetCapacityObjectId": target_capacity_id.upper(),
                    "workspacesToAssign": batch,
                }
            ]
        }

        _base_api(
            request="/v1.0/myorg/admin/capacities/AssignWorkspaces",
            method="post",
            payload=payload,
        )

    print(
        f"{icons.green_dot} The workspaces have been assigned to the '{target_capacity}' capacity. A total of {len(workspaces)} were moved."
    )


@log
def unassign_workspaces_from_capacity(
    workspaces: str | List[str] | UUID | List[UUID],
):
    """
    Unassigns workspace(s) from their capacity.

    This is a wrapper function for the following API: `Admin - Capacities UnassignWorkspacesFromCapacity <https://learn.microsoft.com/rest/api/power-bi/admin/capacities-unassign-workspaces-from-capacity>`_.

    Parameters
    ----------
    workspaces : str | List[str] | uuid.UUID | List[uuid.UUID]
        The workspace name(s) or ID(s).
    """
    if isinstance(workspaces, str):
        workspaces = [workspaces]

    dfW = list_workspaces()
    workspacesIds = dfW[dfW["Name"].isin(workspaces)]["Id"].tolist()
    workspacesIds = workspacesIds + dfW[dfW["Id"].isin(workspaces)]["Id"].tolist()

    if len(workspacesIds) != len(workspaces):
        raise ValueError(
            f"{icons.red_dot} Some of the workspaces provided are not valid."
        )

    payload = {"workspacesToUnassign": workspacesIds}
    _base_api(
        request="/v1.0/myorg/admin/capacities/UnassignWorkspaces",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} A total of {len(workspacesIds)} workspaces have been unassigned."
    )


@log
def list_modified_workspaces(
    modified_since: Optional[str] = None,
    exclude_inactive_workspaces: Optional[bool] = False,
    exclude_personal_workspaces: Optional[bool] = False,
) -> pd.DataFrame:
    """
    Gets a list of workspace IDs in the organization.

    This is a wrapper function for the following API: `Admin - WorkspaceInfo GetModifiedWorkspaces <https://learn.microsoft.com/rest/api/power-bi/admin/workspace-info-get-modified-workspaces>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    modified_since : str, default=None
        Last modified date (must be in ISO compliant UTC format). Example: "2024-11-02T05:51:30" or "2024-11-02T05:51:30.0000000Z".
    exclude_inactive_workspaces : bool, default=False
        Whether to exclude inactive workspaces.
    exclude_personal_workspaces : bool, default=False
        Whether to exclude personal workspaces.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of workspace IDs in the organization.
    """
    params = {}
    url = "/v1.0/myorg/admin/workspaces/modified"

    if modified_since is not None:
        modified_since_dt = dtparser(modified_since)
        params["modifiedSince"] = (
            f"{modified_since_dt.isoformat(timespec='microseconds')}0Z"
        )

    if exclude_inactive_workspaces is not None:
        params["excludeInActiveWorkspaces"] = exclude_inactive_workspaces

    if exclude_personal_workspaces is not None:
        params["excludePersonalWorkspaces"] = exclude_personal_workspaces

    url = _build_url(url, params)
    response = _base_api(request=url, client="fabric_sp")

    df = pd.DataFrame(response.json()).rename(columns={"id": "Workspace Id"})

    return df


@log
def list_workspace_access_details(
    workspace: Optional[Union[str, UUID]] = None,
) -> pd.DataFrame:
    """
    Shows a list of users (including groups and Service Principals) that have access to the specified workspace.

    This is a wrapper function for the following API: `Workspaces - List Workspace Access Details <https://learn.microsoft.com/rest/api/fabric/admin/workspaces/list-workspace-access-details>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of users (including groups and Service Principals) that have access to the specified workspace.
    """
    (workspace_name, workspace_id) = _resolve_workspace_name_and_id(workspace)

    columns = {
        "User Id": "string",
        "User Name": "string",
        "User Type": "string",
        "Workspace Name": "string",
        "Workspace Id": "string",
        "Workspace Role": "string",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(
        request=f"/v1/admin/workspaces/{workspace_id}/users", client="fabric_sp"
    )

    rows = []
    for v in response.json().get("accessDetails", []):
        rows.append(
            {
                "User Id": v.get("principal", {}).get("id"),
                "User Name": v.get("principal", {}).get("displayName"),
                "User Type": v.get("principal", {}).get("type"),
                "Workspace Name": workspace_name,
                "Workspace Id": workspace_id,
                "Workspace Role": v.get("workspaceAccessDetails", {}).get(
                    "workspaceRole"
                ),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def _resolve_workspace_name(workspace_id: Optional[UUID] = None) -> str:
    from sempy_labs._helper_functions import _get_fabric_context_setting
    from sempy.fabric.exceptions import FabricHTTPException

    if workspace_id is None:
        workspace_id = _get_fabric_context_setting(name="trident.workspace.id")

    try:
        workspace_name = (
            _base_api(
                request=f"/v1/admin/workspaces/{workspace_id}", client="fabric_sp"
            )
            .json()
            .get("name")
        )
    except FabricHTTPException:
        raise ValueError(
            f"{icons.red_dot} The '{workspace_id}' workspace was not found."
        )
    return workspace_name


@log
def _resolve_workspace_name_and_id(
    workspace: str | UUID,
) -> Tuple[str, UUID]:

    from sempy_labs._helper_functions import _get_fabric_context_setting

    if workspace is None:
        workspace_id = _get_fabric_context_setting(name="trident.workspace.id")
        workspace_name = _resolve_workspace_name(workspace_id)
    elif _is_valid_uuid(workspace):
        workspace_id = workspace
        workspace_name = _resolve_workspace_name(workspace_id)
    else:
        dfW = list_workspaces(workspace=workspace)
        if not dfW.empty:
            workspace_name = dfW["Name"].iloc[0]
            workspace_id = dfW["Id"].iloc[0]
        else:
            raise ValueError(
                f"{icons.red_dot} The '{workspace}' workspace was not found."
            )

    return workspace_name, workspace_id


@log
def list_workspace_users(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows a list of users that have access to the specified workspace.

    This is a wrapper function for the following API: `Admin - Groups GetGroupUsersAsAdmin <https://learn.microsoft.com/rest/api/power-bi/admin/groups-get-group-users-as-admin>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of users that have access to the specified workspace.
    """

    (workspace_name, workspace_id) = _resolve_workspace_name_and_id(workspace)

    columns = {
        "User Name": "string",
        "Email Address": "string",
        "Group User Access Right": "string",
        "Identifier": "string",
        "Graph Id": "string",
        "Principal Type": "string",
    }

    df = _create_dataframe(columns=columns)

    url = f"/v1.0/myorg/admin/groups/{workspace_id}/users"
    response = _base_api(request=url, client="fabric_sp")

    rows = []
    for v in response.json().get("value", []):
        rows.append(
            {
                "User Name": v.get("displayName"),
                "Email Address": v.get("emailAddress"),
                "Group User Access Right": v.get("groupUserAccessRight"),
                "Identifier": v.get("identifier"),
                "Graph Id": v.get("graphId"),
                "Principal Type": v.get("principalType"),
            }
        )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df



from .._helper_functions import (
    _base_api,
    _is_valid_uuid,
)
from uuid import UUID
from sempy_labs._tags import list_tags
import sempy_labs._icons as icons
from typing import List
from sempy._utils._log import log


@log
def resolve_tag_id(tag: str | UUID):

    if _is_valid_uuid(tag):
        tag_id = tag
    else:
        df = list_tags()
        df[df["Tag Name"] == tag]
        if df.empty:
            raise ValueError(f"{icons.red_dot} The '{tag}' tag does not exist.")
        tag_id = df.iloc[0]["Tag Id"]

    return tag_id


@log
def create_tags(tags: str | List[str]):
    """
    Creates a new tag or tags.

    This is a wrapper function for the following API: `Tags - Bulk Create Tags <https://learn.microsoft.com/rest/api/fabric/admin/tags/bulk-create-tags>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    tags : str | List[str]
        The name of the tag or tags to create.
    """

    if isinstance(tags, str):
        tags = [tags]

    # Check the length of the tags
    for tag in tags:
        if len(tag) > 40:
            raise ValueError(
                f"{icons.red_dot} The '{tag}' tag name is too long. It must be 40 characters or less."
            )

    # Check if the tags already exist
    df = list_tags()
    existing_names = df["Tag Name"].tolist()
    existing_ids = df["Tag Id"].tolist()

    available_tags = [
        tag for tag in tags if tag not in existing_names and tag not in existing_ids
    ]
    unavailable_tags = [
        tag for tag in tags if tag in existing_names or tag in existing_ids
    ]

    print(f"{icons.warning} The following tags already exist: {unavailable_tags}")
    if not available_tags:
        print(f"{icons.info} No new tags to create.")
        return

    payload = [{"displayName": name} for name in available_tags]

    for tag in tags:
        _base_api(
            request="/v1/admin/bulkCreateTags",
            client="fabric_sp",
            method="post",
            payload=payload,
            status_codes=201,
        )

    print(f"{icons.green_dot} The '{available_tags}' tag(s) have been created.")


@log
def delete_tag(tag: str | UUID):
    """
    Deletes a tag.

    This is a wrapper function for the following API: `Tags - Delete Tag <https://learn.microsoft.com/rest/api/fabric/admin/tags/delete-tag>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    tag : str | uuid.UUID
        The name or ID of the tag to delete.
    """

    tag_id = resolve_tag_id(tag)

    _base_api(request=f"/v1/admin/tags/{tag_id}", client="fabric_sp", method="delete")

    print(f"{icons.green_dot} The '{tag}' tag has been deleted.")


@log
def update_tag(name: str, tag: str | UUID):
    """
    Updates the name of a tag.

    This is a wrapper function for the following API: `Tags - Update Tag <https://learn.microsoft.com/rest/api/fabric/admin/tags/update-tag>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name : str
        The new name of the tag.
    tag : str | uuid.UUID
        The name or ID of the tag to update.
    """

    tag_id = resolve_tag_id(tag)

    _base_api(
        request=f"/v1/admin/tags/{tag_id}",
        client="fabric_sp",
        method="patch",
        payload={"displayName": name},
    )

    print(f"{icons.green_dot} The '{tag}' tag has been renamed to '{name}'.")



from ._items import (
    list_mirrored_databases,
    get_mirrored_database_definition,
    create_mirrored_database,
    delete_mirrored_database,
    get_mirroring_status,
    get_tables_mirroring_status,
    stop_mirroring,
    start_mirroring,
    update_mirrored_database_definition,
)

__all__ = [
    "list_mirrored_databases",
    "get_mirrored_database_definition",
    "create_mirrored_database",
    "delete_mirrored_database",
    "get_mirroring_status",
    "get_tables_mirroring_status",
    "stop_mirroring",
    "start_mirroring",
    "update_mirrored_database_definition",
]



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_id,
    _base_api,
    _create_dataframe,
    delete_item,
    create_item,
)
from uuid import UUID
from sempy._utils._log import log


@log
def list_kql_querysets(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the KQL querysets within a workspace.

    This is a wrapper function for the following API: `Items - List KQL Querysets <https://learn.microsoft.com/rest/api/fabric/kqlqueryset/items/list-kql-querysets>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the KQL querysets within a workspace.
    """

    columns = {
        "KQL Queryset Name": "string",
        "KQL Queryset Id": "string",
        "Description": "string",
    }
    df = _create_dataframe(columns=columns)

    workspace_id = resolve_workspace_id(workspace)

    responses = _base_api(
        request=f"v1/workspaces/{workspace_id}/kqlQuerysets",
        uses_pagination=True,
        client="fabric_sp",
    )

    rows = []
    for r in responses:
        for v in r.get("value", []):
            rows.append(
                {
                    "KQL Queryset Name": v.get("displayName"),
                    "KQL Queryset Id": v.get("id"),
                    "Description": v.get("description"),
                }
            )

    if rows:
        df = pd.DataFrame(rows, columns=list(columns.keys()))

    return df


@log
def create_kql_queryset(
    name: str, description: Optional[str] = None, workspace: Optional[str | UUID] = None
):
    """
    Creates a KQL queryset.

    This is a wrapper function for the following API: `Items - Create KQL Queryset <https://learn.microsoft.com/rest/api/fabric/kqlqueryset/items/create-kql-queryset>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    name: str
        Name of the KQL queryset.
    description : str, default=None
        A description of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    create_item(
        name=name, description=description, type="KQLQueryset", workspace=workspace
    )


@log
def delete_kql_queryset(
    kql_queryset: str | UUID, workspace: Optional[str | UUID] = None
):
    """
    Deletes a KQL queryset.

    This is a wrapper function for the following API: `Items - Delete KQL Queryset <https://learn.microsoft.com/rest/api/fabric/kqlqueryset/items/delete-kql-queryset>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    kql_queryset: str | uuid.UUID
        Name or ID of the KQL queryset.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=kql_queryset, type="KQLQueryset", workspace=workspace)



from typing import Optional, Literal
from uuid import UUID
from sempy_labs._helper_functions import (
    _base_api,
    resolve_workspace_name_and_id,
    resolve_item_name_and_id,
)
import sempy_labs._icons as icons
from sempy._utils._log import log


@log
def takeover_item_ownership(
    item: str | UUID,
    type: Literal["Report", "SemanticModel"],
    workspace: Optional[str | UUID] = None,
):
    """
    Takes over ownership of a Report or Semantic Model.

    Parameters
    ----------
    item : str | uuid.UUID
        The name or UUID of the Report or Semantic Model to take over.
    type : typing.Literal["Report", "SemanticModel"]
        The type of item to take over. Must be either 'Report' or 'SemanticModel'.
    workspace : Optional[str | uuid.UUID], default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    if type not in ["Report", "SemanticModel"]:
        raise ValueError(
            f"{icons.red_dot} Type must be either 'Report' or 'SemanticModel'."
        )

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (item_name, item_id) = resolve_item_name_and_id(item, type, workspace_id)

    map = {
        "Report": "reports",
        "SemanticModel": "datasets",
    }

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/{map.get(type)}/{item_id}/Default.TakeOver",
        method="post",
        client="fabric_sp",
    )

    print(
        f"{icons.green_dot} Successfully took over the '{item_name}' {type} in workspace '{workspace_name}'."
    )



import pandas as pd
from typing import Optional
from uuid import UUID
from sempy._utils._log import log
import sempy_labs.mirrored_warehouse as mw


@log
def list_mirrored_warehouses(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the mirrored warehouses within a workspace.

    This is a wrapper function for the following API: `Items - List Mirrored Warehouses <https://learn.microsoft.com/rest/api/fabric/mirroredwarehouse/items/list-mirrored-warehouses>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the mirrored warehouses within a workspace.
    """

    return mw.list_mirrored_warehouses(workspace=workspace)



from ._generate_shared_expression import generate_shared_expression
from ._directlake_schema_compare import direct_lake_schema_compare
from ._directlake_schema_sync import direct_lake_schema_sync
from ._dl_helper import (
    check_fallback_reason,
    generate_direct_lake_semantic_model,
    get_direct_lake_source,
)
from ._get_shared_expression import get_shared_expression
from ._guardrails import (
    get_direct_lake_guardrails,
    get_sku_size,
    get_directlake_guardrails_for_sku,
)
from ._list_directlake_model_calc_tables import (
    list_direct_lake_model_calc_tables,
)
from ._show_unsupported_directlake_objects import (
    show_unsupported_direct_lake_objects,
)
from ._update_directlake_model_lakehouse_connection import (
    update_direct_lake_model_lakehouse_connection,
    update_direct_lake_model_connection,
)
from ._update_directlake_partition_entity import (
    update_direct_lake_partition_entity,
    add_table_to_direct_lake_semantic_model,
)
from ._warm_cache import (
    warm_direct_lake_cache_isresident,
    warm_direct_lake_cache_perspective,
)
from ._autosync import set_autosync
from ._sources import get_direct_lake_sources

__all__ = [
    "generate_shared_expression",
    "direct_lake_schema_compare",
    "direct_lake_schema_sync",
    "check_fallback_reason",
    "get_shared_expression",
    "get_direct_lake_guardrails",
    "get_sku_size",
    "get_directlake_guardrails_for_sku",
    "list_direct_lake_model_calc_tables",
    "show_unsupported_direct_lake_objects",
    "update_direct_lake_model_lakehouse_connection",
    "update_direct_lake_partition_entity",
    "warm_direct_lake_cache_isresident",
    "warm_direct_lake_cache_perspective",
    "add_table_to_direct_lake_semantic_model",
    "generate_direct_lake_semantic_model",
    "get_direct_lake_source",
    "update_direct_lake_model_connection",
    "set_autosync",
    "get_direct_lake_sources",
]
